"""In this module, we provide built-in support for many array types."""

# pyright: reportMissingImports=none
from __future__ import annotations

import json
import logging
import sys
import warnings
from abc import ABC, abstractmethod
from collections.abc import Hashable, Mapping, Sequence
from functools import cached_property
from typing import TYPE_CHECKING, Any, ClassVar, Generic, Protocol, TypeVar

import numpy as np
import numpy.typing as npt
from psygnal import Signal

from ._ring_buffer import RingBuffer

if TYPE_CHECKING:
    from collections.abc import Container, Iterator
    from typing import Any, TypeAlias, TypeGuard

    import dask.array.core as da
    import numpy.typing as npt
    import pydantic_core
    import pyopencl.array as cl_array
    import sparse
    import tensorstore as ts
    import torch
    import xarray as xr
    from pydantic import GetCoreSchemaHandler

    Index: TypeAlias = int | slice


class SupportsIndexing(Protocol):
    def __getitem__(self, key: Index | tuple[Index, ...]) -> npt.ArrayLike: ...
    @property
    def shape(self) -> tuple[int, ...]: ...


ArrayT = TypeVar("ArrayT")
NPArrayLike = TypeVar("NPArrayLike", bound=SupportsIndexing)
_T = TypeVar("_T", bound=type)
logger = logging.getLogger("ndv")


def _recurse_subclasses(cls: _T) -> Iterator[_T]:
    for subclass in cls.__subclasses__():
        yield subclass
        yield from _recurse_subclasses(subclass)


class DataWrapper(Generic[ArrayT], ABC):
    """Interface for wrapping different array-like data types.

    [`DataWrapper.create()`][ndv.DataWrapper.create] is a factory method that returns a
    `DataWrapper` instance for the given data type. If your datastore type is not
    supported, you may implement a new `DataWrapper` subclass to handle your data type.
    To do this, import and subclass `DataWrapper`, and (minimally) implement the
    supports and isel methods. Ensure that your class is imported before the
    `DataWrapper.create` method is called, and it will be automatically detected and
    used to wrap your data.

    This base class provides basic support for numpy-like array types.  If the data
    supports __getitem__ and shape attributes, it will work.  If the data does not
    support __getitem__, the subclass MUST implement the `isel` method.  If the data
    does not have a `shape` attribute, the subclass MUST implement the `dims` and
    `coords` properties.
    """

    # Order in which subclasses are checked for support.
    # Lower numbers are checked first, and the first supporting subclass is used.
    # Default is 50, and fallback to numpy-like duckarray is 100.
    # Subclasses can override this to change the priority in which they are checked
    PRIORITY: ClassVar[int] = 50
    # These names will be checked when looking for a channel axis
    COMMON_CHANNEL_NAMES: ClassVar[Container[str]] = ("channel", "ch", "c")
    COMMON_Z_AXIS_NAMES: ClassVar[Container[str]] = ("z", "depth", "focus")

    # Maximum dimension size consider when guessing the channel axis
    MAX_CHANNELS: ClassVar[int] = 16

    dims_changed = Signal()
    """Signal to emit when the dimensions of the data change.

    NOTE: It is up to data wrappers, or even end-users to emit this signal when the
    dimensions/shape of the wrapped _data object changes.
    """
    data_changed = Signal()
    """Signal emitted when the data changes.

    NOTE: It is up to data wrappers, or even end-users to emit this signal when the
    data object changes.  We do not currently use object proxies to spy on mutation
    of the underlying data.
    """

    def __init__(self, data: ArrayT) -> None:
        self._data = data
        if not hasattr(self._data, "__getitem__") and "isel" not in type(self).__dict__:
            raise NotImplementedError(
                "DataWrapper subclass MUST implement `isel` method if data does not "
                "support __getitem__."
            )

        has_shape = hasattr(self._data, "shape") and isinstance(self._data.shape, tuple)
        has_methods = "dims" in type(self).__dict__ and "coords" in type(self).__dict__
        if not has_shape and not has_methods:
            raise NotImplementedError(
                "DataWrapper subclass MUST implement `dims` and `coords` properties"
                " if data does not have a `shape` attribute or if the shape is not "
                "a tuple."
            )
        self.dims_changed.connect(self.clear_cache)

    # ----------------------------- Mandatory methods -----------------------------

    @classmethod
    @abstractmethod
    def supports(cls, obj: Any) -> TypeGuard[Any]:
        """Return True if this wrapper can handle the given object.

        Any exceptions raised by this method will be suppressed, so it is safe to
        directly import necessary dependencies without a try/except block.
        """

    @property
    def dims(self) -> tuple[Hashable, ...]:
        """Return the dimension labels for the data."""
        if hasattr(self._data, "dims"):
            return tuple(self._data.dims)
        # type ignore is asserted in the __init__ method
        return tuple(range(len(self._data.shape)))  # type: ignore [attr-defined]

    @property
    def coords(self) -> Mapping[Hashable, Sequence]:
        """Return the coordinates for the data."""
        if hasattr(self._data, "coords") and isinstance(self._data.coords, Mapping):
            return self._data.coords

        dims = self.dims
        # type ignore is asserted in the __init__ method
        return {i: range(s) for i, s in zip(dims, self._data.shape, strict=False)}  # type: ignore [attr-defined]

    def isel(self, index: Mapping[int, int | slice]) -> np.ndarray:
        """Return a slice of the data as a numpy array.

        `index` will look like (e.g.) `{0: slice(0, 10), 1: 5}`.
        The default implementation converts the index to a tuple of the same length as
        the self.dims, populating missing keys with `slice(None)`, and then slices the
        data array using __getitem__.
        """
        idx = tuple(index.get(k, slice(None)) for k in range(len(self.dims)))
        # this type ignore is asserted in the __init__ method
        # if the data does not support __getitem__, then the DataWrapper subclass will
        # fail to initialize
        return self._asarray(self._data[idx])  # type: ignore [index]

    def _asarray(self, data: Any) -> np.ndarray:
        """Convert data to a numpy array."""
        return np.asarray(data)

    def save_as_zarr(self, path: str) -> None:
        raise NotImplementedError("Saving as zarr is not supported for this data type")

    @property
    def significant_bits(self) -> int | None:
        """Number of significant bits per sample, if known from metadata."""
        return None

    @property
    def dtype(self) -> np.dtype:
        """Return the dtype for the data."""
        try:
            return np.dtype(self._data.dtype)  # type: ignore[attr-defined]
        except AttributeError as e:
            raise NotImplementedError(
                "`dtype` property not properly implemented for DataWrapper of type: "
                f"{type(self)}"
            ) from e

    # -----------------------------

    @classmethod
    def create(cls, data: ArrayT) -> DataWrapper[ArrayT]:
        """Create a DataWrapper instance for the given data.

        This method will detect all subclasses of DataWrapper and check them in order of
        their `PRIORITY` class variable. The first subclass that
        [`supports`][ndv.DataWrapper.supports] the given data will be used to wrap it.

        !!! tip

            This means that you can subclass DataWrapper to handle new data types.
            Just make sure that your subclass is imported before calling `create`.

        If no subclasses support the data, a `NotImplementedError` is raised.

        If an instance of `DataWrapper` is passed in, it will be returned as-is.
        """
        if isinstance(data, DataWrapper):
            return data

        # check subclasses for support
        # This allows users to define their own DataWrapper subclasses which will
        # be automatically detected (assuming they have been imported by this point)
        for subclass in sorted(_recurse_subclasses(cls), key=lambda x: x.PRIORITY):
            try:
                if subclass.supports(data):
                    logger.debug("Using %s to wrap %s", subclass.__name__, type(data))
                    return subclass(data)
            except Exception as e:
                warnings.warn(
                    f"Error checking DataWrapper subclass {subclass.__name__}: {e}",
                    RuntimeWarning,
                    stacklevel=2,
                )
        raise NotImplementedError(f"Don't know how to wrap type {type(data)}")

    @property
    def data(self) -> ArrayT:
        """Return the data being wrapped."""
        return self._data

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source: type, handler: GetCoreSchemaHandler
    ) -> pydantic_core.CoreSchema:
        from pydantic_core import core_schema

        return core_schema.no_info_before_validator_function(
            function=cls.create,
            schema=core_schema.any_schema(),
        )

    def sizes(self) -> Mapping[Hashable, int]:
        """Return the sizes of the dimensions."""
        return {dim: len(self.coords[dim]) for dim in self.dims}

    # these guess_x methods may change in the future to become more agnostic to the
    # dimension name/semantics that they are guessing.

    def guess_channel_axis(self) -> Hashable | None:
        """Return the (best guess) axis name for the channel dimension."""
        # for arrays with labeled dimensions,
        # see if any of the dimensions are named "channel"
        sizes = self.sizes()
        if len(sizes) < 3 or min(sizes.values()) > self.MAX_CHANNELS:
            return None

        for dimkey, val in sizes.items():
            if str(dimkey).lower() in self.COMMON_CHANNEL_NAMES:
                if val <= self.MAX_CHANNELS:
                    return self.normalize_axis_key(dimkey)

        # otherwise use the smallest dimension as the channel axis
        return min(sizes, key=sizes.get)  # type: ignore [arg-type]

    def guess_z_axis(self) -> Hashable | None:
        """Return the (best guess) axis name for the z (3rd spatial) dimension."""
        sizes = self.sizes()
        ch = self.guess_channel_axis()
        for dimkey in sizes:
            if str(dimkey).lower() in self.COMMON_Z_AXIS_NAMES:
                if (normed := self.normalize_axis_key(dimkey)) != ch:
                    return normed

        # otherwise return the LAST axis that is neither in the last two dimensions
        # or the channel axis guess
        return next(
            (self.normalize_axis_key(x) for x in reversed(self.dims[:-2]) if x != ch),
            None,
        )

    def summary_info(self) -> str:
        """Return info label with information about the data."""
        package = getattr(self._data, "__module__", "").split(".")[0]
        info = f"{package}.{getattr(type(self._data), '__qualname__', '')}"

        if sizes := self.sizes():
            # if all of the dimension keys are just integers, omit them from size_str
            if all(isinstance(x, int) for x in sizes):
                size_str = repr(tuple(sizes.values()))
            # otherwise, include the keys in the size_str
            else:
                size_str = ", ".join(f"{k}:{v}" for k, v in sizes.items())
                size_str = f"({size_str})"
            info += f" {size_str}"
        if dtype := getattr(self._data, "dtype", ""):
            info += f", {dtype}"
        if nbytes := getattr(self._data, "nbytes", 0):
            info += f", {_human_readable_size(nbytes)}"
        return info

    @cached_property
    def axis_map(self) -> Mapping[Hashable, int]:
        """Mapping of ALL valid axis keys to normalized, positive integer keys."""
        axis_index: dict[Hashable, int] = {}
        ndims = len(self.dims)
        for i, dim in enumerate(self.dims):
            axis_index[dim] = i  # map dimension label to positive index
            axis_index[i] = i  # map positive integer index to itself
            axis_index[-(ndims - i)] = i  # map negative integer index to positive index
        return axis_index

    def normalize_axis_key(self, axis: Hashable) -> int:
        """Return positive index for `axis` (which can be +/- int or str label)."""
        try:
            return self.axis_map[axis]
        except KeyError as e:
            ndims = len(self.dims)
            if isinstance(axis, int):
                raise IndexError(
                    f"Axis index {axis} out of bounds for data with {ndims} dimensions"
                ) from e
            raise IndexError(f"Axis label {axis} not found in data dimensions") from e

    def channel_names(self, channel_axis: int | None) -> dict[int, str]:
        """Return channel display names from data metadata."""
        if channel_axis is None:
            return {}
        dim = self.dims[channel_axis]
        coords = self.coords.get(dim, None)
        # range coords are not informative, so return empty dict in that case
        # and let resolution logic fallback to generic channel naming
        if coords is None or isinstance(coords, range):
            return {}
        return {
            i: str(v.item() if hasattr(v, "item") else v) for i, v in enumerate(coords)
        }

    def axis_scales(self) -> dict[Hashable, float]:
        """Return per-axis scale factors from coordinate spacing."""
        scales: dict[Hashable, float] = {}
        for dim in self.dims:
            coord = self.coords.get(dim)
            if coord is None or len(coord) < 2:
                continue
            try:
                values = [float(v) for v in coord]
            except (TypeError, ValueError):
                continue
            if any(not np.isfinite(v) for v in values):
                continue
            diffs = [values[i + 1] - values[i] for i in range(len(values) - 1)]
            if all(abs(d - diffs[0]) < 1e-10 for d in diffs):
                scales[dim] = diffs[0]
        return scales

    def clear_cache(self) -> None:
        """Clear any cached properties."""
        if hasattr(self, "axis_map"):
            del self.axis_map


def _human_readable_size(nbytes: float) -> str:
    units = ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
    for unit in units:
        if nbytes < 1024:
            return f"{nbytes:.2f}".rstrip("0").rstrip(".") + unit
        nbytes /= 1024.0
    return f"{nbytes:.2f}YB"  # In case nbytes is extremely large


##########################


class ArrayLikeWrapper(DataWrapper[NPArrayLike]):
    """Wrapper for numpy duck array-like objects.

    The base class is suitable for any object that supports __getitem__ and shape.
    So we just need to implement supports to define the type of object we are wrapping.
    """

    PRIORITY = 100

    @classmethod
    def supports(cls, obj: Any) -> TypeGuard[NPArrayLike]:
        if (
            (
                isinstance(obj, np.ndarray)
                or hasattr(obj, "__array_function__")
                or hasattr(obj, "__array_namespace__")
                or hasattr(obj, "__array__")
            )
            and hasattr(obj, "__getitem__")
            and hasattr(obj, "shape")
        ):
            return True
        return False


class DaskWrapper(DataWrapper["da.Array"]):
    """Wrapper for dask array objects."""

    @classmethod
    def supports(cls, obj: Any) -> TypeGuard[da.Array]:
        if (da := sys.modules.get("dask.array")) and isinstance(obj, da.Array):
            return True
        return False

    def _asarray(self, data: da.Array) -> np.ndarray:
        return np.asarray(data.compute())

    def save_as_zarr(self, path: str) -> None:
        self._data.to_zarr(url=path)


class SparseArrayWrapper(DataWrapper["sparse.Array"]):
    PRIORITY = 50

    @classmethod
    def supports(cls, obj: Any) -> TypeGuard[sparse.COO]:
        if (sparse := sys.modules.get("sparse")) and isinstance(obj, sparse.COO):
            return True
        return False

    def _asarray(self, data: sparse.COO) -> np.ndarray:
        return np.asarray(data.todense())


class CLArrayWrapper(DataWrapper["cl_array.Array"]):
    """Wrapper for pyopencl array objects."""

    PRIORITY = 50

    @classmethod
    def supports(cls, obj: Any) -> TypeGuard[cl_array.Array]:
        if (cl_array := sys.modules.get("pyopencl.array")) and isinstance(
            obj, cl_array.Array
        ):
            return True
        return False

    def _asarray(self, data: cl_array.Array) -> np.ndarray:
        return np.asarray(data.get())


class XarrayWrapper(DataWrapper["xr.DataArray"]):
    """Wrapper for xarray DataArray objects."""

    @classmethod
    def supports(cls, obj: Any) -> TypeGuard[xr.DataArray]:
        if (xr := sys.modules.get("xarray")) and isinstance(obj, xr.DataArray):
            return True
        return False


class TensorstoreWrapper(DataWrapper["ts.TensorStore"]):
    """Wrapper for tensorstore.TensorStore objects."""

    def __init__(self, data: Any) -> None:
        super().__init__(data)

        import tensorstore as ts

        self._ts = ts

        spec = self.data.spec().to_json()
        dims: Sequence[Hashable] | None = None
        self._ts = ts
        if (tform := spec.get("transform")) and ("input_labels" in tform):
            dims = [str(x) for x in tform["input_labels"]]
        elif (
            str(spec.get("driver")).startswith("zarr")
            and (zattrs := self.data.kvstore.read(".zattrs").result().value)
            and isinstance((zattr_dict := json.loads(zattrs)), dict)
            and "_ARRAY_DIMENSIONS" in zattr_dict
        ):
            dims = zattr_dict["_ARRAY_DIMENSIONS"]

        if isinstance(dims, Sequence) and len(dims) == len(self._data.domain):
            self._dims: tuple[Hashable, ...] = tuple(str(x) for x in dims)
            self._data = self.data[ts.d[:].label[self._dims]]
        else:
            self._dims = tuple(range(len(self._data.domain)))
        self._coords: Mapping[Hashable, Sequence] = {
            i: range(s)
            for i, s in zip(self._dims, self._data.domain.shape, strict=False)
        }

    @property
    def dtype(self) -> np.dtype:
        """Return the dtype for the data."""
        return np.dtype(str(self._data.dtype.name))

    @property
    def dims(self) -> tuple[Hashable, ...]:
        """Return the dimension labels for the data."""
        return self._dims

    @property
    def coords(self) -> Mapping[Hashable, Sequence]:
        """Return the coordinates for the data."""
        return self._coords

    def sizes(self) -> Mapping[Hashable, int]:
        return dict(zip(self._dims, self._data.domain.shape, strict=False))

    def isel(self, indexers: Mapping[int, int | slice]) -> np.ndarray:
        if not indexers:
            slc: slice | tuple = slice(None)
        else:
            slc = tuple(
                indexers.get(i, slice(None)) for i in range(len(self._data.shape))
            )
        result = self._data[slc].read().result()
        return np.asarray(result)

    @classmethod
    def supports(cls, obj: Any) -> TypeGuard[ts.TensorStore]:
        if (ts := sys.modules.get("tensorstore")) and isinstance(obj, ts.TensorStore):
            return True
        return False


class TorchTensorWrapper(DataWrapper["torch.Tensor"]):
    """Wrapper for torch tensor objects."""

    @property
    def coords(self) -> Mapping[Hashable, Sequence]:
        dims = self.dims
        return {i: range(s) for i, s in zip(dims, self.data.shape, strict=False)}

    @property
    def dims(self) -> tuple[Hashable, ...]:
        # torch does enforce that len(names) == len(shape), and that names are unique
        # with the exception of None, which is allowed to be repeated
        if hasattr(self.data, "names"):
            names = self.data.names
            return tuple((i if name is None else name) for i, name in enumerate(names))
        return tuple(range(len(self.data.shape)))

    @classmethod
    def supports(cls, obj: Any) -> TypeGuard[torch.Tensor]:
        if (torch := sys.modules.get("torch")) and isinstance(obj, torch.Tensor):
            return True
        return False


class RingBufferWrapper(DataWrapper[RingBuffer]):
    """Wrapper for ring buffer objects."""

    def __init__(
        self,
        max_capacity: int | RingBuffer,
        dtype: npt.DTypeLike | None = None,
        *,
        allow_overwrite: bool = True,
    ):
        if isinstance(max_capacity, RingBuffer):
            if dtype is not None:  # pragma: no cover
                raise ValueError(
                    "Cannot specify dtype when passing an existing RingBuffer."
                )
            self._ring = max_capacity
        else:
            if dtype is None:
                dtype = float
            self._ring = RingBuffer(
                max_capacity=max_capacity, dtype=dtype, allow_overwrite=allow_overwrite
            )
        self._ring.resized.connect(self.dims_changed)
        super().__init__(self._ring)

    @property
    def dims(self) -> tuple[int, ...]:
        """Return the dimensions of the data."""
        return tuple(range(len(self._ring.shape)))

    @property
    def coords(self) -> Mapping:
        """Return the coordinates for the data."""
        shape = self._ring.shape
        return {i: range(s) for i, s in enumerate(shape)}

    @classmethod
    def supports(cls, obj: Any) -> TypeGuard[np.ndarray]:
        if isinstance(obj, RingBuffer):
            return True
        return False

    def append(self, value: npt.ArrayLike) -> None:
        """Append a value to the right end of the buffer."""
        self._ring.append(value)

    def appendleft(self, value: npt.ArrayLike) -> None:
        """Append a value to the left end of the buffer."""
        self._ring.appendleft(value)

    def pop(self) -> np.ndarray:
        """Pop a value from the right end of the buffer."""
        return self._ring.pop()

    def popleft(self) -> np.ndarray:
        """Pop a value from the left end of the buffer."""
        return self._ring.popleft()



"""Wrappers around GUI & graphics frameworks.

Most stuff in this module is not intended for public use, but [`ndv.views.bases`][]
shows the protocol that GUI & graphics classes should implement.
"""

from ._app import (
    CanvasBackend,
    GuiFrontend,
    call_later,
    get_array_canvas_class,
    get_array_view_class,
    get_histogram_canvas_class,
    gui_frontend,
    process_events,
    run_app,
    set_canvas_backend,
    set_gui_backend,
)

__all__ = [
    "CanvasBackend",
    "GuiFrontend",
    "call_later",
    "get_array_canvas_class",
    "get_array_view_class",
    "get_histogram_canvas_class",
    "gui_frontend",
    "process_events",
    "run_app",
    "set_canvas_backend",
    "set_gui_backend",
]



from __future__ import annotations

from abc import abstractmethod
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from psygnal import Signal

from ndv._types import KeyPressEvent
from ndv.models._array_display_model import ChannelMode

from ._view_base import Viewable

if TYPE_CHECKING:
    from collections.abc import Container, Hashable, Mapping, Sequence

    from ndv._types import AxisKey, ChannelKey
    from ndv.models._viewer_model import ArrayViewerModel
    from ndv.views.bases import LUTView


class ArrayView(Viewable):
    """ABC for ND Array viewers widget.

    Currently, this is the "main" widget that contains the array display and
    all the controls for interacting with the array, including sliders, LUTs,
    and histograms.
    """

    currentIndexChanged = Signal()
    resetZoomClicked = Signal()
    histogramRequested = Signal(int)
    ndimToggleRequested = Signal(bool)
    channelModeChanged = Signal(ChannelMode)
    keyPressed = Signal(KeyPressEvent)

    @abstractmethod
    def __init__(
        self,
        canvas_widget: Any,
        viewer_model: ArrayViewerModel,
        **kwargs: Any,
    ) -> None: ...
    @abstractmethod
    def create_sliders(self, coords: Mapping[Hashable, Sequence]) -> None: ...
    @abstractmethod
    def current_index(self) -> Mapping[AxisKey, int | slice]: ...
    @abstractmethod
    def set_current_index(self, value: Mapping[AxisKey, int | slice]) -> None: ...

    @abstractmethod
    def visible_axes(self) -> Sequence[AxisKey]: ...
    @abstractmethod
    def set_visible_axes(self, axes: Sequence[AxisKey]) -> None: ...

    @abstractmethod
    def set_channel_mode(self, mode: ChannelMode) -> None: ...
    @abstractmethod
    def set_data_info(self, data_info: str) -> None: ...
    @abstractmethod
    def set_hover_info(self, hover_info: str) -> None: ...
    @abstractmethod
    def hide_sliders(
        self, axes_to_hide: Container[Hashable], *, show_remainder: bool = ...
    ) -> None: ...
    @abstractmethod
    def add_lut_view(self, key: ChannelKey) -> LUTView: ...
    @abstractmethod
    def remove_lut_view(self, view: LUTView) -> None: ...

    def add_histogram(self, channel: ChannelKey, widget: Any) -> None:
        raise NotImplementedError

    def remove_histogram(self, widget: Any) -> None:
        raise NotImplementedError



"""Tests pertaining to Pygfx components"""

from importlib.util import find_spec

import pytest

if not find_spec("pygfx"):
    pytest.skip(
        "Skipping Pygfx tests as Pygfx is not installed", allow_module_level=True
    )



"""Numpy Ring Buffer.

Vendored from https://github.com/eric-wieser/numpy_ringbuffer
April 27, 2025
(and subsequently updated).

MIT License

Copyright (c) 2016 Eric Wieser

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import annotations

import warnings
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, cast, overload

import numpy as np
from psygnal import Signal

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Any, SupportsIndex

    import numpy.typing as npt


class RingBuffer(Sequence):
    """Ring buffer structure with a given capacity and element type.

    Parameters
    ----------
    max_capacity: int
        The maximum capacity of the ring buffer.
    dtype: npt.DTypeLike
        Desired type (and shape) of individual buffer elements.
        This is passed to `np.empty`, so it can be any
        [dtype-like object](https://numpy.org/doc/stable/reference/arrays.dtypes.html).
        Common scenarios will be:
            - a fixed dtype (e.g. `int`, `np.uint8`, `'u2'`, `np.dtype('f4')`)
            - a `(fixed_dtype, shape)` tuple (e.g. `('uint16', (512, 512))`)
    allow_overwrite: bool
        If false, throw an IndexError when trying to append to an already full
        buffer.
    create_buffer: Callable[[int, npt.DTypeLike], npt.NDArray]
        A callable that creates the underlying array.
        May be used to customize the initialization of the array.  Defaults to
        `np.empty`.

    Notes
    -----
    Vendored from [numpy-ringbuffer](https://github.com/eric-wieser/numpy_ringbuffer),
    by Eric Wieser (MIT License).  And updated with typing and signals.
    """

    resized = Signal(int)

    def __init__(
        self,
        max_capacity: int,
        dtype: npt.DTypeLike = float,
        *,
        allow_overwrite: bool = True,
        create_buffer: Callable[[int, npt.DTypeLike], npt.NDArray] = np.empty,
    ) -> None:
        self._arr = create_buffer(max_capacity, dtype)
        self._left_index = 0
        self._right_index = 0
        self._capacity = max_capacity
        self._allow_overwrite = allow_overwrite

    # -------------------- Properties --------------------

    @property
    def is_full(self) -> bool:
        """True if there is no more space in the buffer."""
        return len(self) == self._capacity

    @property
    def dtype(self) -> np.dtype:
        """Return the dtype of the buffer."""
        return self._arr.dtype

    @property
    def shape(self) -> tuple[int, ...]:
        """Return the shape of the valid buffer (excluding unused space)."""
        return (len(self), *self._arr.shape[1:])

    # these mirror methods from deque
    @property
    def maxlen(self) -> int:
        """Return the maximum capacity of the buffer."""
        return self._capacity

    # -------------------- Methods --------------------

    def append(self, value: npt.ArrayLike) -> None:
        """Append a value to the right end of the buffer."""
        if was_full := self.is_full:
            if not self._allow_overwrite:
                raise IndexError("append to a full RingBuffer with overwrite disabled")
            elif not len(self):
                return
            else:
                self._left_index += 1

        self._arr[self._right_index % self._capacity] = value
        self._right_index += 1
        self._fix_indices()
        if not was_full:
            self.resized.emit(len(self))

    def appendleft(self, value: npt.ArrayLike) -> None:
        """Append a value to the left end of the buffer."""
        if was_full := self.is_full:
            if not self._allow_overwrite:
                raise IndexError("append to a full RingBuffer with overwrite disabled")
            elif not len(self):
                return
            else:
                self._right_index -= 1

        self._left_index -= 1
        self._fix_indices()
        self._arr[self._left_index] = value
        if not was_full:
            self.resized.emit(len(self))

    def pop(self) -> np.ndarray:
        """Pop a value from the right end of the buffer."""
        if len(self) == 0:
            raise IndexError("pop from an empty RingBuffer")
        self._right_index -= 1
        self._fix_indices()
        res = cast("np.ndarray", self._arr[self._right_index % self._capacity])
        self.resized.emit(len(self))
        return res

    def popleft(self) -> np.ndarray:
        """Pop a value from the left end of the buffer."""
        if len(self) == 0:
            raise IndexError("pop from an empty RingBuffer")
        res = cast("np.ndarray", self._arr[self._left_index])
        self._left_index += 1
        self._fix_indices()
        self.resized.emit(len(self))
        return res

    def extend(self, values: npt.ArrayLike) -> None:
        """Extend the buffer with the given values."""
        values = np.asarray(values)
        lv = len(values)
        if len(self) + lv > self._capacity:
            if not self._allow_overwrite:
                raise IndexError(
                    "Extending a RingBuffer such that it would overflow, "
                    "with overwrite disabled."
                )
            elif not len(self):
                return
        if lv >= self._capacity:
            # wipe the entire array! - this may not be threadsafe
            self._arr[...] = values[-self._capacity :]
            self._right_index = self._capacity
            self._left_index = 0
            return

        was_full = self.is_full
        ri = self._right_index % self._capacity
        sl1 = np.s_[ri : min(ri + lv, self._capacity)]
        sl2 = np.s_[: max(ri + lv - self._capacity, 0)]
        self._arr[sl1] = values[: sl1.stop - sl1.start]
        self._arr[sl2] = values[sl1.stop - sl1.start :]
        self._right_index += lv

        self._left_index = max(self._left_index, self._right_index - self._capacity)
        self._fix_indices()
        if not was_full:
            self.resized.emit(len(self))

    def extendleft(self, values: npt.ArrayLike) -> None:
        """Prepend the buffer with the given values."""
        values = np.asarray(values)
        lv = len(values)
        if len(self) + lv > self._capacity:
            if not self._allow_overwrite:
                raise IndexError(
                    "Extending a RingBuffer such that it would overflow, "
                    "with overwrite disabled"
                )
            elif not len(self):
                return
        if lv >= self._capacity:
            # wipe the entire array! - this may not be threadsafe
            self._arr[...] = values[: self._capacity]
            self._right_index = self._capacity
            self._left_index = 0
            return

        was_full = self.is_full
        self._left_index -= lv
        self._fix_indices()
        li = self._left_index
        sl1 = np.s_[li : min(li + lv, self._capacity)]
        sl2 = np.s_[: max(li + lv - self._capacity, 0)]
        self._arr[sl1] = values[: sl1.stop - sl1.start]
        self._arr[sl2] = values[sl1.stop - sl1.start :]

        self._right_index = min(self._right_index, self._left_index + self._capacity)
        if not was_full:
            self.resized.emit(len(self))

    # numpy compatibility
    def __array__(
        self, dtype: npt.DTypeLike | None = None, copy: bool | None = None
    ) -> np.ndarray:
        if copy is False:
            warnings.warn(
                "`copy=False` isn't supported. A copy is always created.",
                RuntimeWarning,
                stacklevel=2,
            )
        return np.asarray(self._unwrap(), dtype=dtype)

    # implement Sequence methods
    def __len__(self) -> int:
        """Return the number of valid elements in the buffer."""
        return self._right_index - self._left_index

    @overload  # type: ignore [override]
    def __getitem__(self, key: SupportsIndex) -> Any: ...
    @overload
    def __getitem__(self, key: Any, /) -> np.ndarray: ...
    def __getitem__(self, key: Any) -> np.ndarray | Any:
        """Index into the buffer.

        This supports both simple and fancy indexing.
        """
        # handle simple (b[1]) and basic (b[np.array([1, 2, 3])]) fancy indexing quickly
        if not isinstance(key, tuple):
            item_arr = np.asarray(key)
            if issubclass(item_arr.dtype.type, np.integer):
                # Map negative indices to positive ones
                item_arr = np.where(item_arr < 0, item_arr + len(self), item_arr)
                # Map indices to the range of the buffer
                item_arr = (item_arr + self._left_index) % self._capacity
                return self._arr[item_arr]

        # for everything else, get it right at the expense of efficiency
        return self._unwrap()[key]

    def __iter__(self) -> Iterator[Any]:
        # this is comparable in speed to using itertools.chain
        return iter(self._unwrap())

    def __repr__(self) -> str:
        """Return a string representation of the buffer."""
        return f"<{self.__class__.__name__} of {np.asarray(self)!r}>"

    def _unwrap(self) -> np.ndarray:
        """Copy the data from this buffer into unwrapped form."""
        return np.concatenate(
            (
                self._arr[self._left_index : min(self._right_index, self._capacity)],
                self._arr[: max(self._right_index - self._capacity, 0)],
            )
        )

    def _fix_indices(self) -> None:
        """Enforce our invariant that 0 <= self._left_index < self._capacity."""
        if self._left_index >= self._capacity:
            self._left_index -= self._capacity
            self._right_index -= self._capacity
        elif self._left_index < 0:
            self._left_index += self._capacity
            self._right_index += self._capacity



import numpy as np
import pygfx
from qtpy.QtWidgets import QApplication
from wgpu.gui.qt import QWgpuCanvas

app = QApplication([])
canvas = QWgpuCanvas(size=(512, 512))
renderer = pygfx.renderers.WgpuRenderer(canvas)

# create data
data = np.random.rand(512, 512).astype(np.float32)

# create scene and camera
scene = pygfx.Scene()
camera = pygfx.OrthographicCamera(*data.shape)
camera.local.position = data.shape[0] / 2, data.shape[1] / 2, 0
controller = pygfx.PanZoomController(camera, register_events=renderer)

# add image
scene.add(
    pygfx.Image(
        pygfx.Geometry(grid=pygfx.Texture(data, dim=2)),
        pygfx.ImageBasicMaterial(depth_test=False),
    )
)

canvas.request_draw(lambda: renderer.render(scene, camera))
app.exec()



from __future__ import annotations

import time
import warnings
from contextlib import suppress
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import psygnal
from qtpy.QtCore import QObject, QPoint, QSize, Qt, Signal  # type: ignore[attr-defined]
from qtpy.QtGui import QCursor, QFontDatabase, QMouseEvent, QMovie
from qtpy.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDoubleSpinBox,
    QFormLayout,
    QFrame,
    QGraphicsOpacityEffect,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLayout,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QSplitter,
    QVBoxLayout,
    QWidget,
)
from superqt import (
    QCollapsible,
    QLabeledRangeSlider,
    QLabeledSlider,
)
from superqt.cmap import QColormapComboBox
from superqt.iconify import QIconifyIcon
from superqt.utils import ensure_main_thread, signals_blocked

from ndv._types import AxisKey
from ndv.models._array_display_model import ChannelMode
from ndv.models._lut_model import ClimPolicy, ClimsManual, ClimsPercentile
from ndv.models._viewer_model import ArrayViewerModel, InteractionMode
from ndv.views.bases import ArrayView, LUTView

if TYPE_CHECKING:
    from collections.abc import Container, Hashable, Mapping, Sequence

    import cmap
    from psygnal import EmissionInfo
    from qtpy.QtGui import QIcon

    from ndv._types import AxisKey, ChannelKey
    from ndv.views.bases._graphics._canvas import HistogramCanvas
    from ndv.views.bases._graphics._canvas_elements import (
        CanvasElement,
        RectangularROIHandle,
    )

SLIDER_STYLE = """
QSlider::groove:horizontal {
    height: 15px;
    background: qlineargradient(
        x1:0, y1:0, x2:0, y2:1,
        stop:0 rgba(128, 128, 128, 0.25),
        stop:1 rgba(128, 128, 128, 0.1)
    );
    border-radius: 3px;
}

QSlider::handle:horizontal {
    width: 38px;
    background: #999999;
    border-radius: 3px;
}

QSlider::sub-page:horizontal {
    background: qlineargradient(
        x1:0, y1:0, x2:0, y2:1,
        stop:0 rgba(100, 100, 100, 0.25),
        stop:1 rgba(100, 100, 100, 0.1)
    );
}

QLabel { font-size: 12px; }

QRangeSlider { qproperty-barColor: qlineargradient(
        x1:0, y1:0, x2:0, y2:1,
        stop:0 rgba(100, 80, 120, 0.2),
        stop:1 rgba(100, 80, 120, 0.4)
    )}
"""


class _CmapCombo(QColormapComboBox):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent, allow_user_colormaps=True, add_colormap_text="Add...")
        self.setMinimumSize(140, 21)
        # self.setStyleSheet("background-color: transparent;")

    def showPopup(self) -> None:
        super().showPopup()
        popup = self.findChild(QFrame)
        popup.setMinimumWidth(self.width() + 100)
        popup.move(popup.x(), popup.y() - self.height() - popup.height())

    # TODO: upstream me
    def setCurrentColormap(self, color: cmap.Colormap) -> None:
        """Adds the color to the QComboBox and selects it."""
        idx = 0
        for idx in range(self.count()):
            if item := self.itemColormap(idx):
                if item.name == color.name:
                    # cmap_ is already here - just select it
                    self.setCurrentIndex(idx)
                    return

        # cmap_ not in the combo box - add it!
        self.addColormap(color)
        # then, select it!
        # NB: "Add..." was at idx, now it's at idx+1 and cmap_ is at idx
        self.setCurrentIndex(idx)


class _QSpinner(QLabel):
    SPIN_GIF = str(Path(__file__).parent.parent / "_resources" / "spin.gif")

    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)
        size = QSize(18, 18)
        mov = QMovie(self.SPIN_GIF, parent=self)
        self.setFixedSize(size)
        mov.setSpeed(150)
        mov.start()
        self.setMovie(mov)

        # make semi-transparent
        effect = QGraphicsOpacityEffect(self)
        effect.setOpacity(0.6)
        self.setGraphicsEffect(effect)


class QtPopup(QDialog):
    """A generic popup window."""

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.setModal(False)  # if False, then clicking anywhere else closes it
        self.setWindowFlags(Qt.WindowType.Popup | Qt.WindowType.FramelessWindowHint)

        self.frame = QFrame(self)
        layout = QVBoxLayout(self)
        layout.addWidget(self.frame)
        layout.setContentsMargins(0, 0, 0, 0)

    def show_above_mouse(self, *args: Any) -> None:
        """Show popup dialog above the mouse cursor position."""
        pos = QCursor().pos()  # mouse position
        szhint = self.sizeHint()
        pos -= QPoint(szhint.width() // 2, szhint.height() + 14)
        self.move(pos)
        self.resize(self.sizeHint())
        self.show()


class PlayButton(QPushButton):
    """Just a styled QPushButton that toggles between play and pause icons."""

    fpsChanged = Signal(float)

    PLAY_ICON = "bi:play-fill"
    PAUSE_ICON = "bi:pause-fill"

    def __init__(self, fps: float = 20, parent: QWidget | None = None) -> None:
        icn = QIconifyIcon(self.PLAY_ICON, color="#888888")
        icn.addKey(self.PAUSE_ICON, state=QIconifyIcon.State.On, color="#4580DD")
        super().__init__(icn, "", parent)
        self.spin = QDoubleSpinBox(self)
        self.spin.setRange(0.5, 100)
        self.spin.setValue(fps)
        self.spin.valueChanged.connect(self.fpsChanged)
        self.setCheckable(True)
        self.setFixedSize(14, 18)
        self.setIconSize(QSize(16, 16))
        self.setStyleSheet(r"QPushButton {border: none; padding: 0; margin: 0;}")
        self._popup = QtPopup(self)
        form = QFormLayout(self._popup.frame)
        form.setContentsMargins(6, 6, 6, 6)
        form.addRow("FPS", self.spin)

    def mousePressEvent(self, e: QMouseEvent | None) -> None:
        if e and e.button() == Qt.MouseButton.RightButton:
            self._show_fps_dialog()
        else:
            super().mousePressEvent(e)

    def _show_fps_dialog(self) -> None:
        self._popup.show_above_mouse()


class _DimToggleButton(QPushButton):
    def __init__(self, parent: QWidget | None = None):
        icn = QIconifyIcon("f7:view-2d", color="#333333")
        icn.addKey("f7:view-3d")
        super().__init__(icn, "", parent)
        self.setCheckable(True)


class _QLUTWidget(QWidget):
    def __init__(
        self,
        default_luts: Sequence[Any],
        parent: QWidget | None = None,
    ) -> None:
        super().__init__(parent)

        # -- WIDGETS -- #
        self.visible = QCheckBox()

        self.cmap = _CmapCombo()
        self.cmap.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.cmap.addColormaps(default_luts)

        self.clims = QLabeledRangeSlider(Qt.Orientation.Horizontal)

        WHITE_SS = SLIDER_STYLE + "SliderLabel { font-size: 10px; color: white;}"
        self.clims.setStyleSheet(WHITE_SS)
        self.clims.setHandleLabelPosition(
            QLabeledRangeSlider.LabelPosition.LabelsOnHandle
        )
        self.clims.setEdgeLabelMode(QLabeledRangeSlider.EdgeLabelMode.NoLabel)
        self.clims.setRange(0, 2**16)

        self.auto_clim = QPushButton("Auto")
        self.auto_clim.setMaximumWidth(42)
        self.auto_clim.setCheckable(True)

        add_histogram_icon = QIconifyIcon("foundation:graph-bar")
        self.histogram_btn = QPushButton(add_histogram_icon, "")
        self.histogram_btn.setCheckable(True)

        set_range_icon = QIconifyIcon("fluent:full-screen-maximize-24-filled")
        self.hist_range = QPushButton(set_range_icon, "", self)
        self.hist_range.setVisible(False)

        log_icon = QIconifyIcon("mdi:math-log")
        self.hist_log = QPushButton(log_icon, "", self)
        self.hist_log.setToolTip("log (base 10, count+1)")
        self.hist_log.setCheckable(True)
        self.hist_log.setVisible(False)

        self.upper_tail = QDoubleSpinBox()
        self.upper_tail.setRange(0, 49.9)
        self.upper_tail.setSingleStep(0.1)
        self.upper_tail.setSuffix("%")
        self.upper_tail.setValue(0)

        self.lower_tail = QDoubleSpinBox()
        self.lower_tail.setRange(0, 49.9)
        self.lower_tail.setSingleStep(0.1)
        self.lower_tail.setSuffix("%")
        self.lower_tail.setValue(0)

        self.auto_popup = QtPopup(self)
        form = QFormLayout(self.auto_popup.frame)
        form.setContentsMargins(6, 6, 6, 6)
        form.addRow("Exclude Darkest:", self.lower_tail)
        form.addRow("Exclude Brightest:", self.upper_tail)

        # -- LAYOUT -- #

        # "main" lut controls (always visible)
        self._lut_layout = QHBoxLayout()
        self._lut_layout.setSpacing(5)
        self._lut_layout.setContentsMargins(0, 0, 0, 0)
        self._lut_layout.addWidget(self.visible)
        self._lut_layout.addWidget(self.cmap)
        self._lut_layout.addWidget(self.clims)
        self._lut_layout.addWidget(self.auto_clim)
        self._lut_layout.addWidget(self.histogram_btn)

        # histogram controls go in their own layout
        hist_ctrls_layout = QVBoxLayout()
        # Add a vertical spacer that expands to take up available space
        # This is the key component that pushes everything down
        hist_ctrls_layout.addItem(
            QSpacerItem(0, 0, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding)
        )
        hist_ctrls_layout.addWidget(self.hist_log)
        hist_ctrls_layout.addWidget(self.hist_range)

        # histogram layout contains controls + a histogram (which is added later)
        self.hist_layout = QHBoxLayout()
        self.hist_layout.addLayout(hist_ctrls_layout)

        # Overall layout
        self._layout = QVBoxLayout(self)
        self._layout.setSpacing(0)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.addLayout(self._lut_layout)
        self._layout.addLayout(self.hist_layout)


class QLUTView(LUTView):
    # NB: In practice this will be a ChannelKey but Unions not allowed here.
    histogramRequested = psygnal.Signal(object)

    def __init__(
        self,
        channel: ChannelKey = None,
        default_luts: Sequence[Any] = ("gray", "green", "magenta"),
    ) -> None:
        super().__init__()
        self._qwidget = _QLUTWidget(default_luts)
        self._channel = channel
        self.histogram: HistogramCanvas | None = None
        # TODO: use emit_fast
        self._qwidget.histogram_btn.toggled.connect(self._on_q_histogram_toggled)
        self._qwidget.hist_log.toggled.connect(self._on_log_btn_toggled)
        self._qwidget.hist_range.clicked.connect(self._on_set_histogram_range_clicked)
        self._qwidget.visible.toggled.connect(self._on_q_visibility_changed)
        self._qwidget.cmap.currentColormapChanged.connect(self._on_q_cmap_changed)
        self._qwidget.clims.valueChanged.connect(self._on_q_clims_changed)
        self._qwidget.auto_clim.toggled.connect(self._on_q_auto_changed)
        self._qwidget.lower_tail.valueChanged.connect(self._on_q_tails_changed)
        self._qwidget.upper_tail.valueChanged.connect(self._on_q_tails_changed)

        self._qwidget.auto_clim.setContextMenuPolicy(
            Qt.ContextMenuPolicy.CustomContextMenu
        )
        self._qwidget.auto_clim.customContextMenuRequested.connect(
            self._show_autoscale_popup
        )

    def frontend_widget(self) -> QWidget:
        return self._qwidget

    def set_channel_name(self, name: str) -> None:
        self._qwidget.visible.setText(name)
        parent = self._qwidget.parent()
        while parent is not None:
            if isinstance(parent, _QArrayViewer):
                parent._align_lut_names()
                break
            parent = parent.parent()

    def set_clim_policy(self, policy: ClimPolicy) -> None:
        with signals_blocked(self._qwidget.auto_clim):
            self._qwidget.auto_clim.setChecked(not policy.is_manual)
        if isinstance(policy, ClimsPercentile):
            self._qwidget.lower_tail.setValue(policy.min_percentile)
            self._qwidget.upper_tail.setValue(100 - policy.max_percentile)

    def set_colormap(self, cmap: cmap.Colormap) -> None:
        self._qwidget.cmap.setCurrentColormap(cmap)

    def set_clims(self, clims: tuple[float, float]) -> None:
        # block self._qwidget._clims, otherwise autoscale will be forced off
        with signals_blocked(self._qwidget.clims):
            self._qwidget.clims.setValue(clims)

    def set_clim_bounds(
        self,
        bounds: tuple[float | None, float | None] = (None, None),
    ) -> None:
        # block self._qwidget._clims, otherwise autoscale will be forced off
        mi = 0 if bounds[0] is None else int(bounds[0])
        ma = 65535 if bounds[1] is None else int(bounds[1])
        with signals_blocked(self._qwidget.clims):
            self._qwidget.clims.setRange(mi, ma)

    def set_gamma(self, gamma: float) -> None:
        pass

    def set_channel_visible(self, visible: bool) -> None:
        self._qwidget.visible.setChecked(visible)

    def set_visible(self, visible: bool) -> None:
        self._qwidget.setVisible(visible)

    def close(self) -> None:
        self._qwidget.close()

    def _on_q_visibility_changed(self, visible: bool) -> None:
        if self._model:
            self._model.visible = visible

    def _on_q_cmap_changed(self, cmap: cmap.Colormap) -> None:
        if self._model:
            self._model.cmap = cmap

    def _on_q_clims_changed(self, clims: tuple[float, float]) -> None:
        if self._model:
            self._model.clims = ClimsManual(min=clims[0], max=clims[1])

    def _on_q_auto_changed(self, autoscale: bool) -> None:
        if self._model:
            if autoscale:
                lower_tail = self._qwidget.lower_tail.value()
                upper_tail = self._qwidget.upper_tail.value()
                self._model.clims = ClimsPercentile(
                    min_percentile=lower_tail, max_percentile=100 - upper_tail
                )
            else:
                clims = self._qwidget.clims.value()
                self._model.clims = ClimsManual(min=clims[0], max=clims[1])

    def _on_q_tails_changed(self, value: float) -> None:
        self._on_q_auto_changed(self._qwidget.auto_clim.isChecked())

    def _on_q_histogram_toggled(self, toggled: bool) -> None:
        # Recursively show/hide hist_layout
        self._set_layout_visibility(toggled, self._qwidget.hist_layout)
        # Request histogram if not created yet
        if self.histogram is None:
            self.histogramRequested.emit(self._channel)

    def _set_layout_visibility(self, toggled: bool, layout: QLayout) -> None:
        for i in range(layout.count()):
            item = layout.itemAt(i)
            if item is None:
                continue
            if wdg := item.widget():
                wdg.setVisible(toggled)
            elif child_layout := item.layout():
                self._set_layout_visibility(toggled, child_layout)

    def _on_log_btn_toggled(self, toggled: bool) -> None:
        if hist := self.histogram:
            hist.set_log_base(10 if toggled else None)

    def _on_set_histogram_range_clicked(self) -> None:
        self._qwidget.hist_log.setChecked(False)
        if hist := self.histogram:
            hist.set_range()

    def _add_histogram(self, histogram: HistogramCanvas) -> None:
        # Add widget to view
        self.histogram = histogram
        widget = cast("QWidget", histogram.frontend_widget())
        # Resize widget to a respectable size
        widget.setFixedHeight(100)
        widget.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self._qwidget.hist_layout.addWidget(widget, 1)

    def _show_autoscale_popup(self) -> None:
        self._qwidget.auto_popup.show_above_mouse()


class QRGBView(QLUTView):
    def __init__(self, channel: ChannelKey = None) -> None:
        super().__init__(channel)
        # Hide the cmap selector
        self._qwidget.cmap.setVisible(False)
        # Insert a new label
        self._label = QLabel("RGB")
        self._qwidget._lut_layout.insertWidget(1, self._label)


class ROIButton(QPushButton):
    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)
        self.setCheckable(True)
        self.setToolTip("Add ROI")
        self.setIcon(QIconifyIcon("mdi:vector-rectangle"))


class DimRow(QObject):
    def __init__(
        self, axis: AxisKey, _coords: Sequence, parent: QObject | None
    ) -> None:
        super().__init__(parent)
        self.slider = QLabeledSlider(Qt.Orientation.Horizontal)
        self.index_label = self.slider._label
        self.play_btn = PlayButton()
        self.play_btn.fpsChanged.connect(self.set_fps)
        self.play_btn.toggled.connect(self.set_animated)
        self.label = QLabel(str(axis))
        self.out_of = QLabel(f"/ {len(_coords) - 1}")
        self.out_of.setStyleSheet("margin: 0 0 1px 0;")  # hack

        self._timer_id: int | None = None
        self._play_start_time: float = 0.0
        self._play_start_frame: int = 0
        mono = QFontDatabase.systemFont(QFontDatabase.SystemFont.FixedFont)
        mono.setPointSize(12)
        self.index_label.setFont(mono)
        self.out_of.setFont(mono)

    def set_fps(self, fps: float) -> None:
        self.play_btn.spin.setValue(fps)
        self.set_animated(self.play_btn.isChecked())

    def set_animated(self, animate: bool) -> None:
        if animate:
            if self._timer_id is not None:
                self.killTimer(self._timer_id)
            fps = self.play_btn.spin.value()
            interval = max(1, int(1000 / fps / 2))  # oversample ~2x
            self._play_start_time = time.perf_counter()
            self._play_start_frame = self.slider.value()
            self._timer_id = self.startTimer(interval, Qt.TimerType.PreciseTimer)
            self.play_btn.setChecked(True)
        elif self._timer_id is not None:
            self.killTimer(self._timer_id)
            self._timer_id = None
            self.play_btn.setChecked(False)

    def timerEvent(self, a0: Any) -> None:
        """Handle timer event for play button, move to the next frame."""
        if self._timer_id is None or a0.timerId() != self._timer_id:
            return
        elapsed = time.perf_counter() - self._play_start_time
        fps = self.play_btn.spin.value()
        n_frames = self.slider.maximum() + 1
        target = (self._play_start_frame + int(elapsed * fps)) % n_frames
        if target != self.slider.value():
            self.slider.setValue(target)


class _QDimsSliders(QWidget):
    currentIndexChanged = Signal()

    _rPLAY_BTN = 0
    _rLABEL = 1
    _rSLIDER = 2
    _rINDEX = 3
    _rTOT = 4

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self._sliders: dict[Hashable, QLabeledSlider] = {}
        self._play_btn_visible: bool = True
        self.setStyleSheet(SLIDER_STYLE)

        self._layout = QGridLayout(self)
        self._layout.setSpacing(1)
        self._layout.setContentsMargins(0, 0, 0, 0)

    def create_sliders(self, coords: Mapping[Hashable, Sequence]) -> None:
        """Update sliders with the given coordinate ranges."""
        grid = self._layout
        for axis, _coords in coords.items():
            # Create a slider for axis if necessary
            if axis not in self._sliders:
                row = grid.rowCount()
                dim_row = DimRow(axis, _coords, self)
                dim_row.slider.valueChanged.connect(self.currentIndexChanged)
                if not self._play_btn_visible:
                    dim_row.play_btn.setVisible(False)
                grid.addWidget(dim_row.play_btn, row, self._rPLAY_BTN)
                grid.addWidget(dim_row.label, row, self._rLABEL)
                grid.addWidget(dim_row.slider, row, self._rSLIDER)
                grid.addWidget(
                    dim_row.index_label, row, self._rINDEX, Qt.AlignmentFlag.AlignRight
                )
                grid.addWidget(
                    dim_row.out_of, row, self._rTOT, Qt.AlignmentFlag.AlignLeft
                )
                self._sliders[axis] = dim_row.slider

            # Update axis slider with coordinates
            sld = self._sliders[axis]
            if isinstance(_coords, range):
                step = _coords.step
                start, stop = _coords.start, (_coords.stop - 1)
            else:
                step = 1
                start, stop = 0, len(_coords) - 1
            sld.setSingleStep(step)
            sld.setRange(start, stop)
            self.setRowTotal(sld, stop)

        self.currentIndexChanged.emit()

    def hide_dimensions(
        self, axes_to_hide: Container[Hashable], show_remainder: bool = True
    ) -> None:
        for ax, slider in self._sliders.items():
            if ax in axes_to_hide:
                self.setRowVisible(slider, False)
            elif show_remainder:
                self.setRowVisible(slider, True)

    def _getSliderRow(self, slider: QWidget) -> int | None:
        # Find the row of the given slider
        for r in range(self._layout.rowCount()):
            if item := self._layout.itemAtPosition(r, self._rSLIDER):
                if item.widget() is slider:
                    return r
        return None

    def setRowTotal(self, slider: QWidget, total: int) -> None:
        if (sr := self._getSliderRow(slider)) is None:
            return

        # Update the total and index label widths for the given row
        item = self._layout.itemAtPosition(sr, self._rTOT)
        if item and (label := item.widget()):
            n_chars = len(str(total))
            label = cast("QLabel", label)
            label.setText(f"/ {total}")
            fm = label.fontMetrics()
            wide = fm.horizontalAdvance(f"/ {'8' * n_chars}") + 6
            label.setFixedWidth(wide)
            idx_item = self._layout.itemAtPosition(sr, self._rINDEX)
            if idx_item and (idx_label := idx_item.widget()):
                idx_label.setFixedWidth(wide)

    def setRowVisible(self, slider: QWidget, visible: bool) -> None:
        if (sr := self._getSliderRow(slider)) is None:
            return

        # Toggle visibility of all widgets in the found row
        for c in range(self._layout.columnCount()):
            if c == self._rPLAY_BTN and not self._play_btn_visible:
                # don't show play button when not visible
                continue
            item = self._layout.itemAtPosition(sr, c)
            if item and (widget := item.widget()):
                widget.setVisible(visible)
                if isinstance(widget, PlayButton) and not visible:
                    widget.setChecked(False)

    def current_index(self) -> Mapping[AxisKey, int | slice]:
        """Return the current value of the sliders."""
        return {axis: slider.value() for axis, slider in self._sliders.items()}

    def set_current_index(self, value: Mapping[AxisKey, int | slice]) -> None:
        """Set the current value of the sliders."""
        changed = False
        # only emit signal if the value actually changed
        # NOTE: this may be unnecessary, since usually the only thing calling
        # set_current_index is the controller, which already knows the value
        # however, we use this method directly in testing and it's nice to ensure.
        with signals_blocked(self):
            for axis, val in value.items():
                if isinstance(val, slice):
                    raise NotImplementedError("Slices are not supported yet")
                if slider := self._sliders.get(axis):
                    if slider.value() != val:
                        changed = True
                        slider.setValue(val)
                else:  # pragma: no cover
                    warnings.warn(f"Axis {axis} not found in sliders", stacklevel=2)
        if changed:
            self.currentIndexChanged.emit()

    def set_play_button_visible(self, visible: bool) -> None:
        """Set the visibility of the play button on all rows."""
        self._play_btn_visible = visible
        for row in range(self._layout.rowCount()):
            item = self._layout.itemAtPosition(row, self._rSLIDER)
            if item and (wdg := item.widget()) and not wdg.isVisible():
                continue  # pragma: no cover, skip if the row is hidden
            item = self._layout.itemAtPosition(row, self._rPLAY_BTN)
            if item and (btn := item.widget()):
                btn.setVisible(visible)

    def stop_animations(self) -> None:
        """Stop all animations on the play buttons."""
        for child in self.findChildren(DimRow):
            child.set_animated(False)


class _UpCollapsible(QCollapsible):
    def __init__(
        self,
        title: str = "",
        parent: QWidget | None = None,
        expandedIcon: QIcon | str | None = "▼",
        collapsedIcon: QIcon | str | None = "▲",
    ):
        super().__init__(title, parent, expandedIcon, collapsedIcon)
        # little hack to make the lut collapsible take up less space
        layout = cast("QVBoxLayout", self.layout())
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        if (
            # look-before-leap on private attribute that may change
            hasattr(self, "_content") and (inner := self._content.layout()) is not None
        ):
            inner.setContentsMargins(0, 4, 0, 0)
            inner.setSpacing(0)

        self.setDuration(100)

        # this is a little hack to allow the buttons on the main view (below)
        # share the same row as the LUT toggle button
        layout.removeWidget(self._toggle_btn)
        self.btn_row = QHBoxLayout()
        self.btn_row.setContentsMargins(0, 0, 0, 0)
        self.btn_row.setSpacing(0)
        self.btn_row.addWidget(self._toggle_btn)
        self.btn_row.addStretch()
        layout.addLayout(self.btn_row)

    def setContent(self, content: QWidget) -> None:
        """Replace central widget (the widget that gets expanded/collapsed)."""
        self._content = content
        # this is different from upstream
        cast("QVBoxLayout", self.layout()).insertWidget(0, self._content)
        self._animation.setTargetObject(content)


# this is a PView ... but that would make a metaclass conflict
class _QArrayViewer(QWidget):
    def __init__(self, canvas_widget: QWidget, parent: QWidget | None = None):
        super().__init__(parent)

        self._canvas_widget = canvas_widget
        self.dims_sliders = _QDimsSliders(self)

        # place to display dataset summary
        self.data_info_label = QLabel("", parent=self)
        # place to display arbitrary text
        self.hover_info_label = QLabel("", self)

        # spinner to indicate progress
        self._progress_spinner = _QSpinner(canvas_widget)
        self._progress_spinner.hide()

        # the button that controls the display mode of the channels
        # not using QEnumComboBox because we want to exclude some values for now
        self.channel_mode_combo = QComboBox(self)
        self.channel_mode_combo.addItems(
            [
                ChannelMode.GRAYSCALE.value,
                ChannelMode.COMPOSITE.value,
                ChannelMode.RGBA.value,
            ]
        )

        # button to reset the zoom of the canvas
        # TODO: unify icons across all the view frontends in a new file
        set_range_icon = QIconifyIcon("fluent:full-screen-maximize-24-filled")
        self.set_range_btn = QPushButton(set_range_icon, "", self)

        # button to draw ROIs
        self._roi_handle: RectangularROIHandle | None = None
        self._selection: CanvasElement | None = None
        self.add_roi_btn = ROIButton()

        self.luts = _UpCollapsible(
            "LUTs",
            parent=self,
            expandedIcon=QIconifyIcon("bi:chevron-up", color="#888888"),
            collapsedIcon=QIconifyIcon("bi:chevron-down", color="#888888"),
        )
        self._btn_layout = self.luts.btn_row
        self._btn_layout.setParent(None)
        self.luts.expand(animate=False)

        # button to change number of displayed dimensions
        self.ndims_btn = _DimToggleButton(self)

        self._btn_layout.addWidget(self.channel_mode_combo)
        self._btn_layout.addWidget(self.ndims_btn)
        self._btn_layout.addWidget(self.add_roi_btn)
        self._btn_layout.addWidget(self.set_range_btn)

        self._btns = QWidget()
        self._btns.setLayout(self._btn_layout)

        # above the canvas
        self._info_widget = QWidget()
        self._info_widget.setFixedHeight(16)
        info = QHBoxLayout(self._info_widget)
        self.data_info_label.setMaximumWidth(400)
        info.setContentsMargins(0, 0, 0, 0)
        info.setSpacing(0)
        info.addWidget(self.data_info_label)
        info.addStretch()
        info.addWidget(self.hover_info_label)

        left = QWidget()
        left_layout = QVBoxLayout(left)
        left_layout.setSpacing(2)
        left_layout.setContentsMargins(0, 0, 0, 0)
        left_layout.addWidget(self._info_widget)
        left_layout.addWidget(canvas_widget, 1)
        left_layout.addWidget(self.dims_sliders)
        left_layout.addWidget(self.luts)
        left_layout.addWidget(self._btns)

        self.splitter = QSplitter(Qt.Orientation.Vertical, self)
        self.splitter.addWidget(left)

        layout = QVBoxLayout(self)
        layout.setSpacing(2)
        layout.setContentsMargins(6, 6, 6, 6)
        layout.addWidget(self.splitter)

    def _align_lut_names(self) -> None:
        """Set matching minimum widths on all LUT name checkboxes."""
        lut_widgets = self.luts.findChildren(_QLUTWidget)
        if not lut_widgets:
            return
        max_w = max(w.visible.sizeHint().width() for w in lut_widgets)
        for w in lut_widgets:
            w.visible.setMinimumWidth(max_w)

    def resizeEvent(self, a0: Any) -> None:
        # position at spinner the top right of the canvas_widget:
        canv, spinner = self._canvas_widget, self._progress_spinner
        pad = 4
        spinner.move(canv.width() - spinner.width() - pad, pad)
        super().resizeEvent(a0)

    def closeEvent(self, a0: Any) -> None:
        with suppress(AttributeError):
            del self._canvas_widget
        super().closeEvent(a0)


class QtArrayView(ArrayView):
    def __init__(
        self,
        canvas_widget: QWidget,
        viewer_model: ArrayViewerModel,
    ) -> None:
        self._viewer_model = viewer_model
        self._qwidget = qwdg = _QArrayViewer(canvas_widget)
        # Mapping of channel key to LUTViews
        self._luts: dict[ChannelKey, QLUTView] = {}
        qwdg.add_roi_btn.toggled.connect(self._on_add_roi_clicked)

        self._viewer_model.events.connect(self._on_viewer_model_event)

        # TODO: use emit_fast
        qwdg.dims_sliders.currentIndexChanged.connect(self.currentIndexChanged.emit)
        qwdg.channel_mode_combo.currentTextChanged.connect(
            self._on_channel_mode_changed
        )
        qwdg.set_range_btn.clicked.connect(self.resetZoomClicked.emit)
        qwdg.ndims_btn.toggled.connect(self._on_ndims_toggled)

        self._visible_axes: Sequence[AxisKey] = []

    def add_lut_view(self, channel: ChannelKey) -> QLUTView:
        view = (
            QRGBView(channel)
            if channel == "RGB"
            else QLUTView(channel, self._viewer_model.default_luts)
        )
        self._luts[channel] = view

        view.histogramRequested.connect(self.histogramRequested)
        self._qwidget.luts.addWidget(view.frontend_widget())
        self._qwidget._align_lut_names()
        return view

    def remove_lut_view(self, view: LUTView) -> None:
        if view not in self._luts.values():
            return
        self._qwidget.luts.removeWidget(cast("QLUTView", view).frontend_widget())

    def _on_channel_mode_changed(self, text: str) -> None:
        self.channelModeChanged.emit(ChannelMode(text))

    def add_histogram(self, channel: ChannelKey, widget: HistogramCanvas) -> None:
        if lut := self._luts.get(channel, None):
            lut._add_histogram(widget)

    def remove_histogram(self, widget: QWidget) -> None:
        widget.setParent(None)
        widget.deleteLater()

    def create_sliders(self, coords: Mapping[Hashable, Sequence]) -> None:
        """Update sliders with the given coordinate ranges."""
        self._qwidget.dims_sliders.create_sliders(coords)

    @ensure_main_thread  # type: ignore[untyped-decorator]
    def hide_sliders(
        self, axes_to_hide: Container[Hashable], *, show_remainder: bool = True
    ) -> None:
        """Hide sliders based on visible axes."""
        self._qwidget.dims_sliders.hide_dimensions(axes_to_hide, show_remainder)

    def current_index(self) -> Mapping[AxisKey, int | slice]:
        """Return the current value of the sliders."""
        return self._qwidget.dims_sliders.current_index()

    def set_current_index(self, value: Mapping[AxisKey, int | slice]) -> None:
        """Set the current value of the sliders."""
        self._qwidget.dims_sliders.set_current_index(value)

    def _on_ndims_toggled(self, is_3d: bool) -> None:
        self._qwidget.dims_sliders.stop_animations()
        self.ndimToggleRequested.emit(is_3d)

    def visible_axes(self) -> Sequence[AxisKey]:
        return self._visible_axes  # no widget to control this yet

    def set_visible_axes(self, axes: Sequence[AxisKey]) -> None:
        self._visible_axes = tuple(axes)
        with signals_blocked(self._qwidget.ndims_btn):
            self._qwidget.ndims_btn.setChecked(len(axes) > 2)

    def set_data_info(self, data_info: str) -> None:
        """Set the data info text, above the canvas."""
        self._qwidget.data_info_label.setText(data_info)

    def set_hover_info(self, hover_info: str) -> None:
        """Set the hover info text, below the canvas."""
        self._qwidget.hover_info_label.setText(hover_info)

    def set_channel_mode(self, mode: ChannelMode) -> None:
        """Set the channel mode button text."""
        self._qwidget.channel_mode_combo.setCurrentText(mode.value)

    def set_visible(self, visible: bool) -> None:
        self._qwidget.setVisible(visible)

    def close(self) -> None:
        self._qwidget.close()

    def frontend_widget(self) -> QWidget:
        return self._qwidget

    def _on_add_roi_clicked(self, checked: bool) -> None:
        self._viewer_model.interaction_mode = (
            InteractionMode.CREATE_ROI if checked else InteractionMode.PAN_ZOOM
        )

    def _on_viewer_model_event(self, info: EmissionInfo) -> None:
        sig_name = info.signal.name
        value = info.args[0]
        if sig_name == "show_progress_spinner":
            self._qwidget._progress_spinner.setVisible(value)
        if sig_name == "interaction_mode":
            # If leaving CanvasMode.CREATE_ROI, uncheck the ROI button
            _new, old = info.args
            if old == InteractionMode.CREATE_ROI:
                self._qwidget.add_roi_btn.setChecked(False)
        elif sig_name == "show_histogram_button":
            for lut in self._luts.values():
                lut._qwidget.histogram_btn.setVisible(value)
        elif sig_name == "show_roi_button":
            self._qwidget.add_roi_btn.setVisible(value)
        elif sig_name == "show_channel_mode_selector":
            self._qwidget.channel_mode_combo.setVisible(value)
        elif sig_name == "show_reset_zoom_button":
            self._qwidget.set_range_btn.setVisible(value)
        elif sig_name == "show_3d_button":
            self._qwidget.ndims_btn.setVisible(value)
        elif sig_name == "show_play_button":
            self._qwidget.dims_sliders.set_play_button_visible(value)
        elif sig_name == "show_data_info":
            self._qwidget._info_widget.setVisible(value)
        elif sig_name == "show_controls":
            # Show or hide the entire controls area (dims sliders + LUTs)
            self._qwidget.luts.setVisible(value)
            self._qwidget.dims_sliders.setVisible(value)
            self._qwidget._btns.setVisible(value)



from __future__ import annotations

import os
from types import MethodType
from typing import TYPE_CHECKING, Any

from jupyter_rfb import RemoteFrameBuffer

from ndv._types import (
    KeyCode,
    KeyMod,
    KeyPressEvent,
    MouseButton,
    MouseMoveEvent,
    MousePressEvent,
    MouseReleaseEvent,
)
from ndv.views.bases._app import NDVApp

if TYPE_CHECKING:
    from collections.abc import Callable

    from ndv.views.bases import ArrayView
    from ndv.views.bases._graphics._mouseable import Mouseable


class JupyterAppWrap(NDVApp):
    """Provider for Jupyter notebooks/lab (NOT ipython)."""

    def is_running(self) -> bool:
        if ipy_shell := self._ipython_shell():
            return bool(ipy_shell.__class__.__name__ == "ZMQInteractiveShell")
        return False

    def create_app(self) -> Any:
        if not self.is_running() and not os.getenv("PYTEST_CURRENT_TEST"):
            # if we got here, it probably means that someone used
            # NDV_GUI_FRONTEND=jupyter without actually being in a jupyter notebook
            # we allow it in tests, but not in normal usage.
            raise RuntimeError(  # pragma: no cover
                "Jupyter is not running a notebook shell.  Cannot create app."
            )

        # No app creation needed...
        # but make sure we can actually import the stuff we need
        import ipywidgets  # noqa: F401
        import jupyter  # noqa: F401

    def array_view_class(self) -> type[ArrayView]:
        from ._array_view import JupyterArrayView

        return JupyterArrayView

    @staticmethod
    def mouse_btn(btn: Any) -> MouseButton:
        if btn == 1:
            return MouseButton.LEFT
        if btn == 2:
            return MouseButton.RIGHT
        if btn == 3:
            return MouseButton.MIDDLE
        return MouseButton.NONE

    def filter_mouse_events(
        self, canvas: Any, receiver: Mouseable
    ) -> Callable[[], None]:
        if not isinstance(canvas, RemoteFrameBuffer):
            raise TypeError(
                f"Expected canvas to be RemoteFrameBuffer, got {type(canvas)}"
            )

        # patch the handle_event from _jupyter_rfb.CanvasBackend
        # to intercept various mouse events.
        super_handle_event = canvas.handle_event
        active_btn: MouseButton = MouseButton.NONE

        def handle_event(self: RemoteFrameBuffer, ev: dict) -> None:
            nonlocal active_btn

            intercepted = False
            etype = ev["event_type"]
            if etype == "pointer_move":
                mme = MouseMoveEvent(x=ev["x"], y=ev["y"], btn=active_btn)
                intercepted |= receiver.on_mouse_move(mme)
                if cursor := receiver.get_cursor(mme):
                    canvas.cursor = cursor.to_jupyter()
                receiver.mouseMoved.emit(mme)
            elif etype == "pointer_down":
                if "button" in ev:
                    active_btn = JupyterAppWrap.mouse_btn(ev["button"])
                else:
                    active_btn = MouseButton.NONE
                mpe = MousePressEvent(x=ev["x"], y=ev["y"], btn=active_btn)
                intercepted |= receiver.on_mouse_press(mpe)
                receiver.mousePressed.emit(mpe)
            elif etype == "double_click":
                # Note that in Jupyter, a double_click event is not a pointer event
                # and as such, we need to handle both press and release. See
                # https://github.com/vispy/jupyter_rfb/blob/62831dd5a87bc19b4fd5f921d802ed21141e61ec/js/lib/widget.js#L270
                btn = JupyterAppWrap.mouse_btn(ev["button"])
                mpe = MousePressEvent(x=ev["x"], y=ev["y"], btn=btn)
                intercepted |= receiver.on_mouse_double_press(mpe)
                receiver.mouseDoublePressed.emit(mpe)
                # Release
                mre = MouseReleaseEvent(x=ev["x"], y=ev["y"], btn=btn)
                intercepted |= receiver.on_mouse_release(mre)
                receiver.mouseReleased.emit(mre)
            elif etype == "pointer_up":
                mre = MouseReleaseEvent(x=ev["x"], y=ev["y"], btn=active_btn)
                active_btn = MouseButton.NONE
                intercepted |= receiver.on_mouse_release(mre)
                receiver.mouseReleased.emit(mre)

            if not intercepted:
                super_handle_event(ev)

        canvas.handle_event = MethodType(handle_event, canvas)
        return lambda: setattr(canvas, "handle_event", super_handle_event)

    def filter_key_events(self, widget: Any, receiver: ArrayView) -> Callable[[], None]:
        # In Jupyter, key events must go through the RemoteFrameBuffer canvas
        # (which uses a hidden <input> element to capture keys), not the
        # ipywidgets container. Walk the widget tree to find it.
        target = _find_rfb(widget)
        if target is None:
            return lambda: None

        super_handle_event = target.handle_event

        def handle_event(self: RemoteFrameBuffer, ev: dict) -> None:
            if ev["event_type"] == "key_down":
                key_str = ev.get("key", "")
                key: KeyCode | str
                if key_str in _JUPYTER_KEY_MAP:
                    key = _JUPYTER_KEY_MAP[key_str]
                elif len(key_str) == 1:
                    key = key_str
                else:
                    super_handle_event(ev)
                    return
                mods = KeyMod.NONE
                if ev.get("shiftKey"):
                    mods |= KeyMod.SHIFT
                if ev.get("ctrlKey"):
                    mods |= KeyMod.CTRL
                if ev.get("altKey"):
                    mods |= KeyMod.ALT
                if ev.get("metaKey"):
                    mods |= KeyMod.META
                receiver.keyPressed.emit(KeyPressEvent(key, mods))
            super_handle_event(ev)

        target.handle_event = MethodType(handle_event, target)
        return lambda: setattr(target, "handle_event", super_handle_event)


_JUPYTER_KEY_MAP: dict[str, KeyCode] = {
    "ArrowUp": KeyCode.UP,
    "ArrowDown": KeyCode.DOWN,
    "ArrowLeft": KeyCode.LEFT,
    "ArrowRight": KeyCode.RIGHT,
    " ": KeyCode.SPACE,
    "Home": KeyCode.HOME,
    "End": KeyCode.END,
}


def _find_rfb(widget: Any) -> RemoteFrameBuffer | None:
    """Walk the ipywidgets tree to find a RemoteFrameBuffer child."""
    if isinstance(widget, RemoteFrameBuffer):
        return widget
    for child in getattr(widget, "children", ()):
        if found := _find_rfb(child):
            return found
    return None



from __future__ import annotations

import os
import sys
import threading
import traceback
from concurrent.futures import Executor, Future, ThreadPoolExecutor
from contextlib import suppress
from functools import cache
from threading import Timer
from typing import TYPE_CHECKING, Any, ClassVar, Literal, cast

if TYPE_CHECKING:
    from collections.abc import Callable
    from types import TracebackType

    import pydevd
    from IPython.core.interactiveshell import InteractiveShell
    from typing_extensions import ParamSpec, TypeVar

    from ndv.views.bases import ArrayView
    from ndv.views.bases._graphics._mouseable import Mouseable

    T = TypeVar("T")
    P = ParamSpec("P")

ENV_IPYTHON_GUI_MAGIC = os.getenv("NDV_IPYTHON_MAGIC", "true").lower()
"""Whether to use %gui magic when running in IPython. Default True."""


class NDVApp:
    """Base class for application wrappers."""

    USE_IPY_MAGIC = ENV_IPYTHON_GUI_MAGIC not in ("0", "false", "no")
    # must be valid key for %gui <magic> in IPython
    IPY_MAGIC_KEY: ClassVar[Literal["qt", "wx", None]] = None

    def create_app(self) -> Any:
        """Create the application instance, if not already created."""
        raise NotImplementedError

    def array_view_class(self) -> type[ArrayView]:
        raise NotImplementedError

    def run(self) -> None:
        """Run the application."""
        pass

    def filter_mouse_events(
        self, canvas: Any, receiver: Mouseable
    ) -> Callable[[], None]:
        """Install mouse event filter on `canvas`, redirecting events to `receiver`."""
        raise NotImplementedError

    def filter_key_events(self, widget: Any, receiver: ArrayView) -> Callable[[], None]:
        """Install key event filter on `widget`, emitting `receiver.keyPressed`."""
        raise NotImplementedError

    def call_in_main_thread(
        self, func: Callable[P, T], *args: P.args, **kwargs: P.kwargs
    ) -> Future[T]:
        """Call `func` in the main gui thread."""
        future: Future[T] = Future()
        future.set_result(func(*args, **kwargs))
        return future

    def get_executor(self) -> Executor:
        """Return an executor for running tasks in the background."""
        return _thread_pool_executor()

    @staticmethod
    def _ipython_shell() -> InteractiveShell | None:
        if (ipy := sys.modules.get("IPython")) and (shell := ipy.get_ipython()):
            return cast("InteractiveShell", shell)
        return None

    def _maybe_enable_ipython_gui(self) -> bool:
        if (
            (ipy_shell := self._ipython_shell())
            and self.USE_IPY_MAGIC
            and (key := self.IPY_MAGIC_KEY)
        ):
            ipy_shell.enable_gui(key)  # type: ignore [no-untyped-call]
            return True
        return False

    @staticmethod
    def _install_excepthook() -> None:
        """Install a custom excepthook that does not raise sys.exit().

        This is necessary to prevent the application from closing when an exception
        is raised.
        """
        if hasattr(sys, "_original_excepthook_"):
            # don't install the excepthook more than once
            return
        sys._original_excepthook_ = sys.excepthook  # type: ignore[attr-defined]
        sys.excepthook = ndv_excepthook

    def process_events(self) -> None:
        """Process events for the application."""
        pass

    def call_later(self, msec: int, func: Callable[[], None]) -> None:
        """Call `func` after `msec` milliseconds."""
        # generic implementation using python threading

        Timer(msec / 1000, func).start()


@cache
def _thread_pool_executor() -> ThreadPoolExecutor:
    return ThreadPoolExecutor(max_workers=2)


# -------------------- Exception handling --------------------
DEBUG_EXCEPTIONS = "NDV_DEBUG_EXCEPTIONS"
"""Whether to drop into a debugger when an exception is raised. Default False."""

EXIT_ON_EXCEPTION = "NDV_EXIT_ON_EXCEPTION"
"""Whether to exit the application when an exception is raised. Default False."""


def _print_exception(
    exc_type: type[BaseException],
    exc_value: BaseException,
    exc_traceback: TracebackType | None,
) -> None:
    try:
        import psygnal
        from rich.console import Console
        from rich.traceback import Traceback

        tb = Traceback.from_exception(
            exc_type, exc_value, exc_traceback, suppress=[psygnal], max_frames=10
        )
        Console(stderr=True).print(tb)
    except ImportError:
        traceback.print_exception(exc_type, value=exc_value, tb=exc_traceback)


def ndv_excepthook(
    exc_type: type[BaseException], exc_value: BaseException, tb: TracebackType | None
) -> None:
    _print_exception(exc_type, exc_value, tb)
    if not tb:
        return

    if (
        (debugpy := sys.modules.get("debugpy"))
        and debugpy.is_client_connected()
        and ("pydevd" in sys.modules)
        and (py_db := _pydevd_debugger())
    ):
        with suppress(Exception):
            thread = threading.current_thread()
            additional_info = py_db.set_additional_thread_info(thread)
            additional_info.is_tracing += 1

            try:
                arg = (exc_type, exc_value, tb)
                py_db.stop_on_unhandled_exception(py_db, thread, additional_info, arg)
            finally:
                additional_info.is_tracing -= 1
    elif os.getenv(DEBUG_EXCEPTIONS) in ("1", "true", "True"):
        # Default to pdb if no better option is available
        import pdb

        pdb.post_mortem(tb)

    if os.getenv(EXIT_ON_EXCEPTION) in ("1", "true", "True"):
        print(f"\n{EXIT_ON_EXCEPTION} is set, exiting.")
        sys.exit(1)


def _pydevd_debugger() -> pydevd.PyDB | None:
    with suppress(Exception):
        import pydevd

        return pydevd.get_global_debugger()
    return None



# /// script
# dependencies = [
#     "ndv[pyqt,vispy]",
#     "xarray",
#     "scipy",
#     "pooch",
# ]
# ///
from __future__ import annotations

try:
    import xarray as xr
except ImportError:
    raise ImportError("Please install xarray[io] to run this example")
import ndv

da = xr.tutorial.open_dataset("air_temperature").air
ndv.imshow(
    da.astype("float32"), default_lut={"cmap": "thermal"}, visible_axes=("lat", "lon")
)



"""Tests pertaining to Qt components"""

import pytest

try:
    # NB: We could use importlib, but we'd have to search for qtpy
    # AND for one of the many bindings (PyQt5, PyQt6, PySide2)
    # This seems easier.
    from qtpy.QtCore import Qt  # noqa: F401
except ImportError:
    pytest.skip("Skipping Qt tests as Qt is not installed", allow_module_level=True)



"""Shared utilities for canvas backends."""

from __future__ import annotations

import logging

import numpy as np

logger = logging.getLogger("ndv")


def downsample_data(
    data: np.ndarray, max_size: int, *, warn: bool = True
) -> tuple[np.ndarray, tuple[int, ...]]:
    """Downsample data so no axis exceeds max_size.

    Returns the (possibly downsampled view) data and the per-axis stride factors.
    """
    factors = tuple(
        int(np.ceil(s / max_size)) if s > max_size else 1 for s in data.shape
    )
    if any(f > 1 for f in factors):
        if warn:
            logger.warning(
                "Data shape %s exceeds max texture dimension (%d) and will be "
                "downsampled for rendering (strides: %s).",
                data.shape,
                max_size,
                factors,
            )
        slices = tuple(slice(None, None, f) for f in factors)
        data = data[slices]
    return data, factors



"""Models for `ndv`."""

from ._array_display_model import ArrayDisplayModel, ChannelMode
from ._base_model import NDVModel
from ._data_wrapper import DataWrapper, RingBufferWrapper
from ._lut_model import (
    ClimPolicy,
    ClimsManual,
    ClimsMinMax,
    ClimsPercentile,
    ClimsStdDev,
    LUTModel,
)
from ._resolve import ResolvedDisplayState
from ._ring_buffer import RingBuffer
from ._roi_model import RectangularROIModel
from ._viewer_model import ArrayViewerModel

__all__ = [
    "ArrayDisplayModel",
    "ArrayViewerModel",
    "ChannelMode",
    "ClimPolicy",
    "ClimsManual",
    "ClimsMinMax",
    "ClimsPercentile",
    "ClimsStdDev",
    "DataWrapper",
    "LUTModel",
    "NDVModel",
    "RectangularROIModel",
    "ResolvedDisplayState",
    "RingBuffer",
    "RingBufferWrapper",
]



from __future__ import annotations

from contextlib import contextmanager
from functools import lru_cache
from typing import TYPE_CHECKING

from vispy.app import Canvas
from vispy.gloo import gl
from vispy.gloo.context import get_current_canvas

if TYPE_CHECKING:
    from collections.abc import Generator


@contextmanager
def _opengl_context() -> Generator[None, None, None]:
    """Assure we are running with a valid OpenGL context.

    Only create a Canvas is one doesn't exist. Creating and closing a
    Canvas causes vispy to process Qt events which can cause problems.
    """
    canvas = Canvas(show=False) if get_current_canvas() is None else None
    try:
        yield
    finally:
        if canvas is not None:
            canvas.close()


@lru_cache
def get_max_texture_sizes() -> tuple[int | None, int | None]:
    """Return the maximum texture sizes for 2D and 3D rendering.

    Returns
    -------
    Tuple[int | None, int | None]
        The max textures sizes for (2d, 3d) rendering.
    """
    with _opengl_context():
        max_size_2d = gl.glGetParameter(gl.GL_MAX_TEXTURE_SIZE)

    if not max_size_2d:
        max_size_2d = None

    # vispy/gloo doesn't provide the GL_MAX_3D_TEXTURE_SIZE location,
    # but it can be found in this list of constants
    # http://pyopengl.sourceforge.net/documentation/pydoc/OpenGL.GL.html
    with _opengl_context():
        GL_MAX_3D_TEXTURE_SIZE = 32883
        max_size_3d = gl.glGetParameter(GL_MAX_3D_TEXTURE_SIZE)

    if not max_size_3d:
        max_size_3d = None

    return max_size_2d, max_size_3d



from __future__ import annotations

import importlib.util
import os
import sys
from enum import Enum
from functools import wraps
from typing import TYPE_CHECKING, Any, Protocol, cast

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator
    from concurrent.futures import Future
    from typing import Literal

    from IPython.core.interactiveshell import InteractiveShell
    from typing_extensions import ParamSpec, TypeVar

    from ndv.views.bases import ArrayCanvas, ArrayView, HistogramCanvas
    from ndv.views.bases._app import NDVApp
    from ndv.views.bases._graphics._mouseable import Mouseable

    T = TypeVar("T")
    P = ParamSpec("P")

GUI_ENV_VAR = "NDV_GUI_FRONTEND"
"""Preferred GUI frontend. If not set, the first available GUI frontend is used."""

CANVAS_ENV_VAR = "NDV_CANVAS_BACKEND"
"""Preferred canvas backend. If not set, the first available canvas backend is used."""

_APP: NDVApp | None = None
"""Singleton instance of the current (GUI) application. Once set it shouldn't change."""


class GuiFrontend(str, Enum):
    """Enum of available GUI frontends.

    Attributes
    ----------
    QT : str
        [PyQt5/PySide2/PyQt6/PySide6](https://doc.qt.io)
    JUPYTER : str
        [Jupyter notebook/lab](https://jupyter.org)
    WX : str
        [wxPython](https://wxpython.org)
    """

    QT = "qt"
    JUPYTER = "jupyter"
    WX = "wx"


class CanvasBackend(str, Enum):
    """Enum of available canvas backends.

    Attributes
    ----------
    VISPY : str
        [Vispy](https://vispy.org)
    PYGFX : str
        [Pygfx](https://github.com/pygfx/pygfx)
    """

    VISPY = "vispy"
    PYGFX = "pygfx"


class CanvasProvider(Protocol):
    @staticmethod
    def is_imported() -> bool: ...
    @staticmethod
    def is_available() -> bool: ...
    @staticmethod
    def array_canvas_class() -> type[ArrayCanvas]: ...
    @staticmethod
    def histogram_canvas_class() -> type[HistogramCanvas]: ...


class VispyProvider(CanvasProvider):
    @staticmethod
    def is_imported() -> bool:
        return "vispy" in sys.modules

    @staticmethod
    def is_available() -> bool:
        return importlib.util.find_spec("vispy") is not None

    @staticmethod
    def array_canvas_class() -> type[ArrayCanvas]:
        from vispy.app import use_app

        from ndv.views._vispy._array_canvas import VispyArrayCanvas

        # these may not be necessary, since we likely have already called
        # create_app by this point and vispy will autodetect that.
        # it's an extra precaution
        _frontend = gui_frontend()
        if _frontend == GuiFrontend.JUPYTER:
            use_app("jupyter_rfb")
        elif _frontend == GuiFrontend.WX:
            use_app("wx")
        elif _frontend == GuiFrontend.QT:
            from qtpy import API_NAME

            use_app(API_NAME.lower())

        return VispyArrayCanvas

    @staticmethod
    def histogram_canvas_class() -> type[HistogramCanvas]:
        from ndv.views._vispy._histogram import VispyHistogramCanvas

        return VispyHistogramCanvas


class PygfxProvider(CanvasProvider):
    @staticmethod
    def is_imported() -> bool:
        return "pygfx" in sys.modules

    @staticmethod
    def is_available() -> bool:
        return importlib.util.find_spec("pygfx") is not None

    @staticmethod
    def array_canvas_class() -> type[ArrayCanvas]:
        from ndv.views._pygfx._array_canvas import GfxArrayCanvas

        return GfxArrayCanvas

    @staticmethod
    def histogram_canvas_class() -> type[HistogramCanvas]:
        from ndv.views._pygfx._histogram import PyGFXHistogramCanvas

        return PyGFXHistogramCanvas


# -------------------- Provider selection --------------------

# list of available GUI frontends and canvas backends, tried in order

GUI_PROVIDERS: dict[GuiFrontend, tuple[str, str]] = {
    GuiFrontend.QT: ("ndv.views._qt._app", "QtAppWrap"),
    GuiFrontend.WX: ("ndv.views._wx._app", "WxAppWrap"),
    GuiFrontend.JUPYTER: ("ndv.views._jupyter._app", "JupyterAppWrap"),
}
MOD_TO_KEY = {mod: key for key, (mod, _) in GUI_PROVIDERS.items()}
CANVAS_PROVIDERS: dict[CanvasBackend, CanvasProvider] = {
    CanvasBackend.VISPY: VispyProvider,
    CanvasBackend.PYGFX: PygfxProvider,
}


def _running_apps() -> Iterator[GuiFrontend]:
    """Return an iterator of running GUI applications."""
    for mod_name in ("PyQt5", "PySide2", "PySide6", "PyQt6"):
        if mod := sys.modules.get(f"{mod_name}.QtWidgets"):
            if (
                qapp := getattr(mod, "QApplication", None)
            ) and qapp.instance() is not None:
                yield GuiFrontend.QT

    if (wx := sys.modules.get("wx")) and wx.App.Get() is not None:
        yield GuiFrontend.WX

    if (ipy := sys.modules.get("IPython")) and (shell := ipy.get_ipython()):
        shell = cast("InteractiveShell", shell)
        if shell.__class__.__name__ == "ZMQInteractiveShell":
            yield GuiFrontend.JUPYTER


def _load_app(module: str, cls_name: str) -> NDVApp:
    mod = importlib.import_module(module)
    cls = getattr(mod, cls_name)
    return cast("NDVApp", cls())


def ndv_app() -> NDVApp:
    """Return the active [`GuiFrontend`][ndv.views.GuiFrontend].

    This is determined first by the `NDV_GUI_FRONTEND` environment variable, after which
    known GUI providers are tried in order until one is found that is either already
    running, or available.
    """
    global _APP
    if _APP is not None:
        return _APP

    running = list(_running_apps())

    # Try 1: Load a frontend explicitly requested by the user
    requested = os.getenv(GUI_ENV_VAR, "").lower()
    valid = {x.value for x in GuiFrontend}
    if requested:
        if requested not in valid:
            raise ValueError(
                f"Invalid GUI frontend: {requested!r}. Valid options: {valid}"
            )
        # ensure the app is created for explicitly requested frontends
        _APP = _load_app(*GUI_PROVIDERS[GuiFrontend(requested)])
        _APP.create_app()
        return _APP

    # Try 2: Utilize an existing, running app
    for key, provider in GUI_PROVIDERS.items():
        if key in running:
            _APP = _load_app(*provider)
            _APP.create_app()
            return _APP

    # Try 3: Load an existing app
    errors: list[tuple[str, BaseException]] = []
    for key, provider in GUI_PROVIDERS.items():
        try:
            _APP = _load_app(*provider)
            _APP.create_app()
            return _APP
        except Exception as e:
            errors.append((key, e))

    raise RuntimeError(  # pragma: no cover
        f"Could not find an appropriate GUI frontend: {valid!r}. Tried:\n\n"
        + "\n".join(f"- {key}: {err}" for key, err in errors)
    )


def set_canvas_backend(backend: Literal["pygfx", "vispy"] | None = None) -> None:
    """Sets the preferred canvas backend. Cannot be set after the GUI is running."""
    if _APP:
        raise RuntimeError("Cannot change the backend once the app is running")
    if backend is None:
        os.environ.pop(CANVAS_ENV_VAR)
    else:
        os.environ[CANVAS_ENV_VAR] = CanvasBackend(backend).value  # validate


def set_gui_backend(backend: Literal["jupyter", "qt", "wx"] | None = None) -> None:
    """Sets the preferred GUI backend. Cannot be set after the GUI is running."""
    if _APP:
        raise RuntimeError("Cannot change the backend once the app is running")
    if backend is None:
        os.environ.pop(GUI_ENV_VAR)
    else:
        os.environ[GUI_ENV_VAR] = GuiFrontend(backend).value  # validate


def gui_frontend() -> GuiFrontend:
    # possibly temporary hack to get the current frontend as an enum, for back-compat
    return MOD_TO_KEY[ndv_app().__module__]


def canvas_backend(requested: str | None = None) -> CanvasBackend:
    """Return the preferred canvas backend.

    This is determined first by the NDV_CANVAS_BACKEND environment variable, after which
    CANVAS_PROVIDERS are tried in order until one is found that is either already
    imported or available
    """
    backend = requested or os.getenv(CANVAS_ENV_VAR, "").lower()

    valid = {x.value for x in CanvasBackend}
    if backend:
        if backend not in valid:
            raise ValueError(
                f"Invalid canvas backend: {backend!r}. Valid options: {valid}"
            )
        return CanvasBackend(backend)

    for key, provider in CANVAS_PROVIDERS.items():
        if provider.is_imported():
            return key
    errors: list[tuple[CanvasBackend, BaseException]] = []
    for key, provider in CANVAS_PROVIDERS.items():
        try:
            if provider.is_available():
                return key
        except Exception as e:
            errors.append((key, e))

    raise RuntimeError(  # pragma: no cover
        f"Could not find an appropriate canvas backend: {valid!r}. Tried:\n\n"
        + "\n".join(f"- {key.value}: {err}" for key, err in errors)
    )


def get_array_view_class() -> type[ArrayView]:
    """Return [`ArrayView`][ndv.views.bases.ArrayView] class for current GUI frontend."""  # noqa: E501
    return ndv_app().array_view_class()


def get_array_canvas_class(backend: str | None = None) -> type[ArrayCanvas]:
    """Return [`ArrayCanvas`][ndv.views.bases.ArrayCanvas] class for current canvas backend."""  # noqa: E501
    _backend = canvas_backend(backend)
    if _backend not in CANVAS_PROVIDERS:  # pragma: no cover
        raise NotImplementedError(f"No canvas backend found for {_backend}")
    return CANVAS_PROVIDERS[_backend].array_canvas_class()


def get_histogram_canvas_class(backend: str | None = None) -> type[HistogramCanvas]:
    """Return [`HistogramCanvas`][ndv.views.bases.HistogramCanvas] class for current canvas backend."""  # noqa: E501
    _backend = canvas_backend(backend)
    if _backend not in CANVAS_PROVIDERS:  # pragma: no cover
        raise NotImplementedError(f"No canvas backend found for {_backend}")
    return CANVAS_PROVIDERS[_backend].histogram_canvas_class()


def filter_mouse_events(canvas: Any, receiver: Mouseable) -> Callable[[], None]:
    """Intercept mouse events on `scene_canvas` and forward them to `receiver`.

    Parameters
    ----------
    canvas : Any
        The front-end canvas widget to intercept mouse events from.
    receiver : Mouseable
        The object to forward mouse events to.

    Returns
    -------
    Callable[[], None]
        A function that can be called to remove the event filter.
    """
    return ndv_app().filter_mouse_events(canvas, receiver)


def filter_key_events(widget: Any, receiver: ArrayView) -> Callable[[], None]:
    """Intercept key events on `widget` and emit `receiver.keyPressed`."""
    return ndv_app().filter_key_events(widget, receiver)


def call_later(msec: int, func: Callable[[], None]) -> None:
    """Call `func` after `msec` milliseconds.

    This can be used to enqueue a function to be called after the current event loop
    iteration.  For example, before calling `run_app()`, to ensure that the event
    loop is running before the function is called.

    Parameters
    ----------
    msec : int
        The number of milliseconds to wait before calling `func`.
    func : Callable[[], None]
        The function to call.
    """
    ndv_app().call_later(msec, func)


def process_events() -> None:
    """Force processing of events for the application."""
    ndv_app().process_events()


def run_app() -> None:
    """Start the active GUI application event loop."""
    ndv_app().run()


def ensure_main_thread(func: Callable[P, T]) -> Callable[P, Future[T]]:
    """Decorator that ensures a function is called in the main thread."""

    @wraps(func)
    def _wrapper(*args: P.args, **kwargs: P.kwargs) -> Future[T]:
        fn = ndv_app().call_in_main_thread
        return fn(func, *args, **kwargs)

    return _wrapper


def submit_task(func: Callable[P, T], *args: P.args, **kwargs: P.kwargs) -> Future[T]:
    """Submit a task to the GUI event loop.

    Returns a `concurrent.futures.Future` or an `asyncio.Future` depending on the
    GUI frontend.
    """
    executor = ndv_app().get_executor()
    return executor.submit(func, *args, **kwargs)



"""Resolve display model + data wrapper into a deterministic display state."""

from __future__ import annotations

import logging
import math
from dataclasses import dataclass, field, fields
from typing import TYPE_CHECKING

from ._array_display_model import ChannelMode

if TYPE_CHECKING:
    from collections.abc import Hashable, Iterable, Mapping

    import numpy as np

    from ndv._types import ChannelKey

    from ._array_display_model import ArrayDisplayModel
    from ._data_wrapper import DataWrapper

logger = logging.getLogger("ndv")


@dataclass(frozen=True, slots=True)
class DataRequest:
    """Request object for data slicing."""

    wrapper: DataWrapper = field(repr=False)
    index: Mapping[int, int | slice]
    visible_axes: tuple[int, ...]
    channel_axis: int | None
    channel_mode: ChannelMode


@dataclass(frozen=True, slots=True)
class DataResponse:
    """Response object containing sliced, display-ready arrays.

    Each array in `data` has its first `n_visible_axes` dimensions ordered
    to match `request.visible_axes` (slowest-varying first).  All non-visible,
    non-channel dimensions are collapsed.

    Dimensionality contract:
    - RGBA mode: each array has `n_visible_axes + 1` dims (trailing RGB/A).
    - All other modes: each array has exactly `n_visible_axes` dims.
    """

    n_visible_axes: int
    data: Mapping[ChannelKey, np.ndarray] = field(repr=False)
    request: DataRequest | None = None


@dataclass(frozen=True, slots=True)
class ResolvedDisplayState:
    """Frozen snapshot of resolved display state (ArrayDisplayModel + DataWrapper).

    At any given time, the viewer's display state is influenced by both the current
    ArrayDisplayModel (e.g. which axes are visible, channel mode, current_index), which
    is mutable and can be updated by the user, and the DataWrapper, which has awareness
    of the underlying data (e.g. dimension names, coordinates, metadata).

    The term "resolution" here refers to the process of taking the user's "intentions",
    as expressed in the `ArrayDisplayModel`, and *resolving* them against the reality of
    the data, as represented by the DataWrapper.  This includes normalizing axis keys to
    positive integers, injecting values derived from the data (e.g. guessing a channel
    axis or inferring coordinates), and computing derived state.

    The output of that resolution process is this `ResolvedDisplayState`, which is a
    deterministic, immutable, hashable snapshot of the display state that can be used
    for diffing and driving the actual data slicing and visualization logic.

    Produced by `resolve()`. Used for diffing in `ArrayViewer._apply_changes()`.
    """

    visible_axes: tuple[int, ...]
    channel_axis: int | None
    channel_mode: ChannelMode
    current_index: dict[int, int | slice]
    data_coords: dict[int, tuple]
    hidden_sliders: frozenset[Hashable]
    summary_info: str
    visible_scales: tuple[float, ...]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ResolvedDisplayState):
            return NotImplemented
        return (
            self.visible_axes == other.visible_axes
            and self.channel_axis == other.channel_axis
            and self.channel_mode == other.channel_mode
            and self.current_index == other.current_index
            and self.data_coords == other.data_coords
            and self.hidden_sliders == other.hidden_sliders
            and self.visible_scales == other.visible_scales
            # summary_info excluded: metadata-only, should not trigger data fetch
        )

    def __hash__(self) -> int:
        return id(self)

    def __rich_repr__(self) -> Iterable[tuple[str, object]]:
        for _field in fields(self):
            if _field.name == "data_coords":
                # return truncated data_coords for readability
                yield (
                    _field.name,
                    {
                        k: v if len(v) <= 5 else ((*v[:2], "...", *v[-2:]))
                        for k, v in getattr(self, _field.name).items()
                    },
                )
            else:
                yield _field.name, getattr(self, _field.name)


EMPTY_STATE = ResolvedDisplayState(
    visible_axes=(),
    channel_axis=None,
    channel_mode=ChannelMode.GRAYSCALE,
    current_index={},
    data_coords={},
    hidden_sliders=frozenset(),
    summary_info="",
    visible_scales=(),
)


def _norm_visible_axes(
    model: ArrayDisplayModel, wrapper: DataWrapper
) -> tuple[int, ...]:
    """Return visible_axes as positive integers.

    Falls back to the last N dims if any axis is stale.
    """
    try:
        return tuple(wrapper.normalize_axis_key(ax) for ax in model.visible_axes)
    except (IndexError, KeyError):
        n = min(len(model.visible_axes), len(wrapper.dims))
        return tuple(range(len(wrapper.dims) - n, len(wrapper.dims)))


def _norm_channel_axis(model: ArrayDisplayModel, wrapper: DataWrapper) -> int | None:
    """Return channel_axis as a positive integer, or None.

    If the model has no channel_axis but is in a multichannel mode,
    guess one from the wrapper (avoiding visible axes).
    """
    if model.channel_axis is not None:
        try:
            return wrapper.normalize_axis_key(model.channel_axis)
        except (IndexError, KeyError):
            return None  # stale channel_axis from previous data?

    if model.channel_mode == ChannelMode.GRAYSCALE:
        return None

    guess = wrapper.guess_channel_axis()
    if guess is None:
        return None

    try:
        normed_guess = wrapper.normalize_axis_key(guess)
    except Exception:
        return None

    # don't use a visible axis as the channel axis
    normed_vis = _norm_visible_axes(model, wrapper)
    if normed_guess in normed_vis:
        return None

    return normed_guess


def _norm_current_index(
    model: ArrayDisplayModel, wrapper: DataWrapper
) -> dict[int, int | slice]:
    """Return current_index with positive integer axis keys.

    Handles duplicate keys (e.g. both 'Z' and 0 mapping to the same axis)
    by giving priority to non-integer keys and logging a warning.
    Does NOT mutate the model.
    """
    output: dict[int, int | slice] = {}
    source_key: dict[int, Hashable] = {}  # tracks which original key wrote each entry

    for key, val in model.current_index.items():
        try:
            normed = wrapper.normalize_axis_key(key)
        except (IndexError, KeyError):
            continue  # stale key from previous data
        if normed in output:
            # prefer named keys (e.g. "Z") over raw integer keys
            is_raw_int = normed == key
            winner = source_key[normed] if is_raw_int else key
            logger.warning(
                "Axis key %r normalized to %r, which is also in current_index. "
                "Using %r value.",
                winner,
                normed,
                winner,
            )
            if is_raw_int:
                continue

        output[normed] = val
        source_key[normed] = key

    # Fill in default (0) for any axis not specified by the model,
    # and clamp integer values to valid coordinate ranges.
    coords = wrapper.coords
    for key in coords:
        ax = wrapper.normalize_axis_key(key)
        if ax not in output:
            output[ax] = 0
        else:
            val = output[ax]
            if isinstance(val, int):
                max_val = len(coords[key]) - 1
                # clamp to valid coordinate range
                output[ax] = max(0, min(val, max_val))

    return output


def _norm_data_coords(wrapper: DataWrapper) -> dict[int, tuple]:
    """Return data coordinates as {positive_int: tuple_of_values}."""
    return {
        wrapper.normalize_axis_key(d): tuple(wrapper.coords[d]) for d in wrapper.dims
    }


def _resolve_visible_scales(
    model: ArrayDisplayModel,
    wrapper: DataWrapper,
    visible_axes: tuple[int, ...],
) -> tuple[float, ...]:
    """Return scale factors aligned with visible_axes.

    Priority: user explicit (model.scales) > data-derived > default (1.0).
    """
    data_scales = wrapper.axis_scales()
    scales: list[float] = []
    for ax in visible_axes:
        # check user model.scales (try both positive int and dim label)
        user_scale = None
        for key in model.scales:
            try:
                if wrapper.normalize_axis_key(key) == ax:
                    user_scale = model.scales[key]
                    break
            except (IndexError, KeyError):
                continue
        if user_scale is not None:
            scales.append(user_scale)
        else:
            # try data-derived: axis_scales uses dim labels as keys
            dim_label = wrapper.dims[ax]
            if dim_label in data_scales:
                scales.append(data_scales[dim_label])
            elif ax in data_scales:
                scales.append(data_scales[ax])
            else:
                scales.append(1.0)
    return tuple(scales)


def _compute_hidden_sliders(
    visible_axes: tuple[int, ...],
    channel_axis: int | None,
    channel_mode: ChannelMode,
    data_coords: dict[int, tuple],
    wrapper: DataWrapper,
) -> frozenset[Hashable]:
    """Compute the set of slider keys to hide."""
    hidden: set[int] = set(visible_axes)
    if channel_mode.is_multichannel() and channel_axis is not None:
        hidden.add(channel_axis)
    # hide singleton axes
    hidden.update(ax for ax, coord in data_coords.items() if len(coord) < 2)
    # include dim names so sliders hide regardless of key form
    return frozenset(hidden | {wrapper.dims[ax] for ax in hidden})


def resolve(model: ArrayDisplayModel, wrapper: DataWrapper) -> ResolvedDisplayState:
    """Pure function: resolve model + wrapper into a frozen display state.

    Does NOT mutate the model.
    """
    visible_axes = _norm_visible_axes(model, wrapper)
    channel_axis = _norm_channel_axis(model, wrapper)
    current_index = _norm_current_index(model, wrapper)
    data_coords = _norm_data_coords(wrapper)

    hidden_sliders = _compute_hidden_sliders(
        visible_axes, channel_axis, model.channel_mode, data_coords, wrapper
    )
    visible_scales = _resolve_visible_scales(model, wrapper, visible_axes)

    return ResolvedDisplayState(
        visible_axes=visible_axes,
        channel_axis=channel_axis,
        channel_mode=model.channel_mode,
        current_index=current_index,
        data_coords=data_coords,
        hidden_sliders=hidden_sliders,
        summary_info=wrapper.summary_info(),
        visible_scales=visible_scales,
    )


def build_slice_requests(
    resolved: ResolvedDisplayState, wrapper: DataWrapper
) -> list[DataRequest]:
    """Build data slice requests from resolved state."""
    requested_slice: dict[int, int | slice] = dict(resolved.current_index)

    for ax in resolved.visible_axes:
        if not isinstance(requested_slice.get(ax), slice):
            requested_slice[ax] = slice(None)

    c_ax = resolved.channel_axis
    if c_ax is not None:
        if resolved.channel_mode.is_multichannel():
            if not isinstance(requested_slice.get(c_ax), slice):
                requested_slice[c_ax] = slice(None)
        else:
            # in non-multichannel mode with channel_axis set,
            # we still want the default_lut behavior
            c_ax = None

    # convert all int indices to single-element slices to preserve dimensions
    for ax, val in requested_slice.items():
        if isinstance(val, int):
            requested_slice[ax] = slice(val, val + 1)

    return [
        DataRequest(
            wrapper=wrapper,
            index=requested_slice,
            visible_axes=resolved.visible_axes,
            channel_axis=c_ax,
            channel_mode=resolved.channel_mode,
        )
    ]


def process_request(req: DataRequest) -> DataResponse:
    """Slice, transpose, and split data into display-ready arrays.

    Visible axes are placed first (matching `req.visible_axes` order) and all
    non-visible, non-channel dimensions are collapsed.  See `DataResponse` for
    the dimensionality contract on the returned arrays.
    """
    data = req.wrapper.isel(req.index)
    vis_ax = req.visible_axes
    if data.size == 0:
        return DataResponse(n_visible_axes=len(vis_ax), data={}, request=req)

    t_dims = vis_ax + tuple(i for i in range(data.ndim) if i not in vis_ax)
    vis_shape = tuple(data.shape[ax] for ax in vis_ax)

    data_response: dict[ChannelKey, np.ndarray] = {}
    ch_ax = req.channel_axis

    if req.channel_mode == ChannelMode.RGBA:
        transposed = data.transpose(*t_dims)
        # Merge all non-visible dims into a single channel dim
        ch_size = math.prod(transposed.shape[len(vis_ax) :])
        data_response["RGB"] = transposed.reshape((*vis_shape, ch_size))
    elif ch_ax is None:
        data_response[None] = data.transpose(*t_dims).reshape(vis_shape)
    else:
        for i in range(data.shape[ch_ax]):
            ch_keepdims = (slice(None),) * ch_ax + (i,) + (None,)
            ch_data = data[ch_keepdims]
            data_response[i] = ch_data.transpose(*t_dims).reshape(vis_shape)

    return DataResponse(n_visible_axes=len(vis_ax), data=data_response, request=req)



# pyright: reportOptionalSubscript=none
from __future__ import annotations

import warnings
from contextlib import suppress
from typing import TYPE_CHECKING, Any, Literal, TypeVar, cast
from weakref import ReferenceType, WeakValueDictionary

import cmap as _cmap
import numpy as np
import vispy
import vispy.color
import vispy.scene
from vispy import scene, visuals
from vispy.util.quaternion import Quaternion

from ndv._types import (
    CursorType,
    MouseButton,
    MouseMoveEvent,
    MousePressEvent,
    MouseReleaseEvent,
)
from ndv.models._viewer_model import ArrayViewerModel, InteractionMode
from ndv.views._app import filter_mouse_events
from ndv.views._util import downsample_data
from ndv.views._vispy._util import get_max_texture_sizes
from ndv.views.bases import ArrayCanvas
from ndv.views.bases._graphics._canvas_elements import (
    CanvasElement,
    ImageHandle,
    RectangularROIHandle,
    ROIMoveMode,
)

if TYPE_CHECKING:
    from collections.abc import Sequence


turn = np.sin(np.pi / 4)
DEFAULT_QUATERNION = Quaternion(turn, turn, 0, 0)


class VispyImageHandle(ImageHandle):
    def __init__(self, visual: visuals.ImageVisual | visuals.VolumeVisual) -> None:
        self._visual = visual
        self._allowed_dims = {2, 3} if isinstance(visual, visuals.ImageVisual) else {3}
        # per-axis downsample strides applied to fit GPU texture limits
        self._downsample_factors: tuple[int, ...] = ()

    def data(self) -> np.ndarray:
        try:
            return self._visual._data  # type: ignore [no-any-return]
        except AttributeError:
            return self._visual._last_data  # type: ignore [no-any-return]

    def set_data(self, data: np.ndarray) -> None:
        if data.ndim not in self._allowed_dims:
            warnings.warn(
                f"Got wrong number of dimensions ({data.ndim}) for vispy "
                f"visual of type {type(self._visual)}.",
                stacklevel=2,
            )
            return

        data, downsample_factors = _downcast_and_downsample(
            data,
            three_d=isinstance(self._visual, visuals.VolumeVisual),
            warn=False,
        )
        self._downsample_factors = downsample_factors
        self._visual.set_data(data)

    def visible(self) -> bool:
        return bool(self._visual.visible)

    def set_visible(self, visible: bool) -> None:
        self._visual.visible = visible

    # TODO: shouldn't be needed
    def can_select(self) -> bool:
        return False

    def selected(self) -> bool:
        return False

    def set_selected(self, selected: bool) -> None:
        raise NotImplementedError("Images cannot be selected")

    def clims(self) -> Any:
        return self._visual.clim

    def set_clims(self, clims: tuple[float, float]) -> None:
        with suppress(ZeroDivisionError):
            self._visual.clim = clims

    def gamma(self) -> float:
        return self._visual.gamma  # type: ignore [no-any-return]

    def set_gamma(self, gamma: float) -> None:
        self._visual.gamma = gamma

    def colormap(self) -> _cmap.Colormap:
        return self._cmap  # FIXME

    def set_colormap(self, cmap: _cmap.Colormap) -> None:
        self._cmap = cmap
        self._visual.cmap = cmap.to_vispy()

    def transform(self) -> np.ndarray:
        raise NotImplementedError

    def set_transform(self, transform: np.ndarray) -> None:
        raise NotImplementedError

    def start_move(self, pos: Sequence[float]) -> None:
        pass

    def move(self, pos: Sequence[float]) -> None:
        pass

    def remove(self) -> None:
        self._visual.parent = None

    def get_cursor(self, event: MouseMoveEvent) -> CursorType | None:
        return None


class VispyRectangle(RectangularROIHandle):
    def __init__(self, parent: Any) -> None:
        self._selected = False
        self._move_mode: ROIMoveMode | None = None
        # NB _move_anchor has different meanings depending on _move_mode
        self._move_anchor: tuple[float, float] = (0, 0)

        # Rectangle handles both fill and border
        self._rect = scene.Rectangle(center=[0, 0], width=1, height=1, parent=parent)
        # NB: Should be greater than image orders BUT NOT handle order
        self._rect.order = 10
        self._rect.interactive = True

        self._handle_data = np.zeros((4, 2))
        self._handle_size = 10  # px
        self._handles = scene.Markers(
            pos=self._handle_data,
            size=self._handle_size,
            scaling="fixed",
            parent=parent,
        )
        # NB: Should be greater than image orders and rect order
        self._handles.order = 100
        self._handles.interactive = True

        self.set_fill(_cmap.Color("transparent"))
        self.set_border(_cmap.Color("yellow"))
        self.set_handles(_cmap.Color("white"))
        self.set_visible(False)

    def _tform(self) -> scene.transforms.BaseTransform:
        return self._rect.transforms.get_transform("canvas", "scene")

    def can_select(self) -> bool:
        return True

    def selected(self) -> bool:
        return self._selected

    def set_selected(self, selected: bool) -> None:
        self._selected = selected
        self._handles.visible = selected and self.visible()

    def set_fill(self, color: _cmap.Color) -> None:
        _vis_color = vispy.color.Color(color.hex)
        # NB We need alpha>0 for selection
        _vis_color.alpha = max(color.alpha, 1e-6)
        self._rect.color = _vis_color

    def set_border(self, color: _cmap.Color) -> None:
        _vis_color = vispy.color.Color(color.hex)
        _vis_color.alpha = color.alpha
        self._rect.border_color = _vis_color

    # TODO: Misleading name?
    def set_handles(self, color: _cmap.Color) -> None:
        _vis_color = vispy.color.Color(color.hex)
        _vis_color.alpha = color.alpha
        self._handles.set_data(face_color=_vis_color)  # pyright: ignore[reportArgumentType]

    def set_bounding_box(
        self, minimum: tuple[float, float], maximum: tuple[float, float]
    ) -> None:
        # NB: Support two diagonal points, not necessarily true min/max
        x1 = float(min(minimum[0], maximum[0]))
        y1 = float(min(minimum[1], maximum[1]))
        x2 = float(max(minimum[0], maximum[0]))
        y2 = float(max(minimum[1], maximum[1]))

        # Update rectangle
        self._rect.center = [(x1 + x2) / 2, (y1 + y2) / 2]
        self._rect.width = max(float(x2 - x1), 1e-30)
        self._rect.height = max(float(y2 - y1), 1e-30)

        # Update handles
        self._handle_data[0] = x1, y1
        self._handle_data[1] = x2, y1
        self._handle_data[2] = x2, y2
        self._handle_data[3] = x1, y2
        self._handles.set_data(pos=self._handle_data)

        # FIXME: These should be called internally upon set_data, right?
        # Looks like https://github.com/vispy/vispy/issues/1899
        self._rect._bounds_changed()
        for v in self._rect._subvisuals:
            v._bounds_changed()
        self._handles._bounds_changed()

    def on_mouse_move(self, event: MouseMoveEvent) -> bool:
        # Convert canvas -> world
        canvas_pos = (event.x, event.y)
        world_pos = self._tform().map(canvas_pos)[:2]
        # moving a handle
        if self._move_mode == ROIMoveMode.HANDLE:
            # The anchor is set to the opposite handle, which never moves.
            self.boundingBoxChanged.emit((world_pos, self._move_anchor))
        # translating the whole roi
        elif self._move_mode == ROIMoveMode.TRANSLATE:
            # The anchor is the mouse position reported in the previous mouse event.
            dx = world_pos[0] - self._move_anchor[0]
            dy = world_pos[1] - self._move_anchor[1]
            # If the mouse moved (dx, dy) between events, the whole ROI needs to be
            # translated that amount.
            new_min = (self._handle_data[0, 0] + dx, self._handle_data[0, 1] + dy)
            new_max = (self._handle_data[2, 0] + dx, self._handle_data[2, 1] + dy)
            self.boundingBoxChanged.emit((new_min, new_max))
            self._move_anchor = world_pos

        return False

    def on_mouse_press(self, event: MousePressEvent) -> bool:
        self.set_selected(True)
        # Convert canvas -> world
        canvas_pos = (event.x, event.y)
        world_pos = self._tform().map(canvas_pos)[:2]
        drag_idx = self._handle_under(world_pos)
        # If a marker is pressed
        if drag_idx is not None:
            opposite_idx = (drag_idx + 2) % 4
            self._move_mode = ROIMoveMode.HANDLE
            self._move_anchor = tuple(self._handle_data[opposite_idx].copy())
        # If the rectangle is pressed
        else:
            self._move_mode = ROIMoveMode.TRANSLATE
            self._move_anchor = world_pos
        return False

    def on_mouse_release(self, event: MouseReleaseEvent) -> bool:
        return False

    def get_cursor(self, event: MouseMoveEvent) -> CursorType | None:
        canvas_pos = (event.x, event.y)
        pos = self._tform().map(canvas_pos)[:2]
        if self._handle_under(pos) is not None:
            center = self._rect.center
            if pos[0] < center[0] and pos[1] < center[1]:
                return CursorType.FDIAG_ARROW
            if pos[0] > center[0] and pos[1] > center[1]:
                return CursorType.FDIAG_ARROW
            return CursorType.BDIAG_ARROW
        return CursorType.ALL_ARROW

    def visible(self) -> bool:
        return bool(self._rect.visible)

    def set_visible(self, visible: bool) -> None:
        self._rect.visible = visible
        self._handles.visible = visible and self.selected()

    def remove(self) -> None:
        self._rect.parent = None
        self._handles.parent = None

    def _handle_under(self, pos: Sequence[float]) -> int | None:
        """Returns an int in [0, 3], or None.

        If an int i, means that the handle at self._positions[i] is at pos.
        If None, there is no handle at pos.
        """
        rad2 = (self._handle_size / 2) ** 2
        for i, p in enumerate(self._handle_data):
            if (p[0] - pos[0]) ** 2 + (p[1] - pos[1]) ** 2 <= rad2:
                return i
        return None


class VispyArrayCanvas(ArrayCanvas):
    """Vispy-based viewer for data.

    All vispy-specific code is encapsulated in this class (and non-vispy canvases
    could be swapped in if needed as long as they implement the same interface).
    """

    def __init__(self, viewer_model: ArrayViewerModel) -> None:
        self._viewer = viewer_model

        self._canvas = scene.SceneCanvas(size=(600, 600))

        # this filter needs to remain in scope for the lifetime of the canvas
        # or mouse events will not be intercepted
        # the returned function can be called to remove the filter, (and it also
        # closes on the event filter and keeps it in scope).
        self._disconnect_mouse_events = filter_mouse_events(self._canvas.native, self)

        self._last_state: dict[Literal[2, 3], Any] = {}

        central_wdg: scene.Widget = self._canvas.central_widget
        self._view: scene.ViewBox = central_wdg.add_view()
        self._ndim: Literal[2, 3] | None = None

        # Maps vispy visuals (scene children) → CanvasElement handles.
        # Entries are added by add_image/add_volume/add_bounding_box.
        # Nobody explicitly removes entries: the controller owns handle
        # lifetimes via ChannelController.handles/lut_views (for images) and
        # _roi_view (for ROIs).  When the controller calls handle.remove()
        # (in _clear_canvas) those refs are dropped, the handle is GC'd, and
        # the WeakValueDictionary entry is automatically removed.
        # NB: a WeakKeyDictionary would create a ref cycle here because
        # each handle (value) holds a strong ref back to its visual (key).
        self._elements = WeakValueDictionary[scene.Node, CanvasElement]()
        self._selection: CanvasElement | None = None
        # Maintain weak reference to last ROI created
        self._last_roi_created: ReferenceType[VispyRectangle] | None = None

    @property
    def _camera(self) -> vispy.scene.cameras.BaseCamera:
        return self._view.camera

    def set_ndim(self, ndim: Literal[2, 3]) -> None:
        """Set the number of dimensions of the displayed data."""
        if ndim == self._ndim:
            return
        elif self._ndim is not None:
            # remember the current state before switching to the new camera
            self._last_state[self._ndim] = self._camera.get_state()

        self._ndim = ndim
        if ndim == 3:
            cam = scene.ArcballCamera(fov=0)
            # this sets the initial view similar to what the panzoom view would have.
            cam._quaternion = DEFAULT_QUATERNION
        else:
            cam = scene.PanZoomCamera(aspect=1, flip=(0, 1))

        # restore the previous state if it exists
        if state := self._last_state.get(ndim):
            cam.set_state(state)
        self._view.camera = cam

    def frontend_widget(self) -> Any:
        return self._canvas.native

    def set_visible(self, visible: bool) -> None: ...

    def close(self) -> None:
        self._disconnect_mouse_events()
        self._canvas.close()

    def refresh(self) -> None:
        self._canvas.update()

    def add_image(self, data: np.ndarray | None = None) -> VispyImageHandle:
        """Add a new Image node to the scene."""
        data, downsample_factors = _downcast_and_downsample(data, three_d=False)
        try:
            img = scene.visuals.Image(
                data, parent=self._view.scene, texture_format="auto"
            )
        except ValueError as e:
            warnings.warn(f"{e}. Falling back to CPUScaledTexture", stacklevel=2)
            img = scene.visuals.Image(data, parent=self._view.scene)

        img.set_gl_state("additive", depth_test=False)
        img.interactive = True
        handle = VispyImageHandle(img)
        handle._downsample_factors = downsample_factors
        self._elements[img] = handle
        if data is not None:
            self.set_range()
        return handle

    def add_volume(self, data: np.ndarray | None = None) -> VispyImageHandle:
        data, downsample_factors = _downcast_and_downsample(data, three_d=True)
        try:
            vol = scene.visuals.Volume(
                data,
                parent=self._view.scene,
                interpolation="nearest",
                texture_format="auto",
            )
        except ValueError as e:
            warnings.warn(f"{e}. Falling back to CPUScaledTexture", stacklevel=2)
            vol = scene.visuals.Volume(
                data, parent=self._view.scene, interpolation="nearest"
            )

        vol.set_gl_state("additive", depth_test=False)
        vol.interactive = True
        handle = VispyImageHandle(vol)
        handle._downsample_factors = downsample_factors
        self._elements[vol] = handle
        if data is not None:
            self.set_range()
        return handle

    def add_bounding_box(self) -> VispyRectangle:
        """Add a new Rectangular ROI node to the scene."""
        roi = VispyRectangle(parent=self._view.scene)
        roi.set_visible(False)
        self._elements[roi._handles] = roi
        self._elements[roi._rect] = roi
        self._last_roi_created = ReferenceType(roi)
        return roi

    def set_scales(self, scales: tuple[float, ...]) -> None:
        """Set per-visible-axis scale factors for rendering."""
        if not scales:
            return
        # scales are in data order (slowest-to-fastest, e.g. ZYX)
        # vispy images use row,col -> y,x mapping, so reverse for XY
        vis_scales = list(reversed(scales))
        # pad to 3 components
        while len(vis_scales) < 3:
            vis_scales.append(1.0)
        sx, sy, sz = vis_scales[0], vis_scales[1], vis_scales[2]
        for handle in self._elements.values():
            if not isinstance(handle, VispyImageHandle):
                continue
            child = handle._visual
            if not isinstance(child, (visuals.ImageVisual, visuals.VolumeVisual)):
                continue
            _sx, _sy, _sz = sx, sy, sz
            # compensate for downsampling so coordinates stay correct
            # factors are in data order; scene order is (x, y, z) = reversed
            factors = handle._downsample_factors
            if factors and any(f > 1 for f in factors):
                rev = list(reversed(factors))
                _sx *= rev[0]
                _sy *= rev[1] if len(rev) > 1 else 1
                _sz *= rev[2] if len(rev) > 2 else 1
            child.transform = vispy.visuals.transforms.STTransform(
                scale=(_sx, _sy, _sz)
            )
        self.set_range()

    def set_range(
        self,
        x: tuple[float, float] | None = None,
        y: tuple[float, float] | None = None,
        z: tuple[float, float] | None = None,
        margin: float = 0.01,
    ) -> None:
        """Update the range of the PanZoomCamera.

        When called with no arguments, the range is set to the full extent of
        the data, accounting for any STTransform scales on image visuals.
        """
        # Compute scaled bounds from image/volume visuals
        has_images = False
        _x = [0.0, 0.0]
        _y = [0.0, 0.0]
        _z = [0.0, 0.0]

        for handle in self._elements.values():
            if isinstance(handle, VispyImageHandle):
                data = handle.data()
                if data is None:
                    continue
                has_images = True
                shape = data.shape
                sx, sy, sz = 1.0, 1.0, 1.0
                tform = handle._visual.transform
                if isinstance(tform, vispy.visuals.transforms.STTransform):
                    sx, sy, sz = tform.scale[:3]
                if len(shape) == 3:
                    # VolumeVisual: data (D,H,W) maps to scene (x=W, y=H, z=D)
                    _x[1] = max(_x[1], shape[2] * sx)
                    _y[1] = max(_y[1], shape[1] * sy)
                    _z[1] = max(_z[1], shape[0] * sz)
                else:
                    # ImageVisual: data (H,W) maps to scene (x=W, y=H)
                    _x[1] = max(_x[1], shape[1] * sx)
                    _y[1] = max(_y[1], shape[0] * sy)

        if not has_images:
            # No image data — fall back to vispy's auto-detection
            self._camera.set_range()
            return

        x = cast("tuple[float, float]", _x) if x is None else x
        y = cast("tuple[float, float]", _y) if y is None else y
        z = cast("tuple[float, float]", _z) if z is None else z

        is_3d = isinstance(self._camera, scene.ArcballCamera)
        if is_3d:
            self._camera._quaternion = DEFAULT_QUATERNION
        self._view.camera.set_range(x=x, y=y, z=z, margin=margin)
        if is_3d:
            # vispy computes scale_factor from the 3D diagonal, which over-zooms
            # the initial top-down view. Override to match the 2D view extent.
            xy_max = max(x[1] - x[0], y[1] - y[0])
            self._camera.scale_factor = xy_max * (1 + 2 * margin)

    def zoom(self, factor: float | tuple, center: tuple[float, float] = (0, 0)) -> None:
        """Zoom in (or out) at the given center (world coordinates)."""
        self._camera.zoom(factor=factor, center=center)

    def canvas_to_world(
        self, pos_xy: tuple[float, float]
    ) -> tuple[float, float, float]:
        """Map XY canvas position (pixels) to XYZ coordinate in world space."""
        return self._view.scene.transform.imap(pos_xy)[:3]  # type: ignore [no-any-return]

    def elements_at(self, pos_xy: tuple[float, float]) -> list[CanvasElement]:
        elements = []
        visuals = self._canvas.visuals_at(pos_xy)
        for vis in visuals:
            if (handle := self._elements.get(vis)) is not None:
                elements.append(handle)
        return elements

    def on_mouse_press(self, event: MousePressEvent) -> bool:
        if self._selection:
            self._selection.set_selected(False)
            self._selection = None
        canvas_pos = (event.x, event.y)
        world_pos = self.canvas_to_world(canvas_pos)[:2]

        # If in CREATE_ROI mode, the new ROI should "start" here.
        if self._viewer.interaction_mode == InteractionMode.CREATE_ROI:
            if self._last_roi_created is None:
                raise ValueError("No ROI to create!")
            if new_roi := self._last_roi_created():
                self._last_roi_created = None
                # HACK: Provide a non-zero starting size so that if the user clicks
                # and immediately releases, it's visible and can be selected again
                _min = world_pos
                _max = (world_pos[0] + 1, world_pos[1] + 1)
                # Put the ROI where the user clicked
                new_roi.boundingBoxChanged.emit((_min, _max))
                # new_roi.set_bounding_box(_min, _max)
                # Make it visible
                new_roi.set_visible(True)
                # Select it so the mouse press event below triggers ROIMoveMode.HANDLE
                # TODO: Make behavior more direct
                new_roi.set_selected(True)

            # All done - exit the mode
            self._viewer.interaction_mode = InteractionMode.PAN_ZOOM

        # Select first selectable object at clicked point
        for vis in self.elements_at(canvas_pos):
            if vis.can_select():
                self._selection = vis
                self._selection.on_mouse_press(event)
                return False

        return False

    def on_mouse_move(self, event: MouseMoveEvent) -> bool:
        if event.btn == MouseButton.LEFT:
            if self._selection and self._selection.selected():
                self._selection.on_mouse_move(event)
                # If we are moving the object, we don't want to move the camera
                return True
        return False

    def on_mouse_release(self, event: MouseReleaseEvent) -> bool:
        if self._selection:
            self._selection.on_mouse_release(event)
        return False

    def get_cursor(self, event: MouseMoveEvent) -> CursorType:
        if self._viewer.interaction_mode == InteractionMode.CREATE_ROI:
            return CursorType.CROSS
        for vis in self.elements_at((event.x, event.y)):
            if cursor := vis.get_cursor(event):
                return cursor
        return CursorType.DEFAULT


T = TypeVar("T", bound="np.ndarray | None")


def _downcast(data: T) -> T:
    """Downcast >32bit data to 32bit."""
    # downcast to 32bit, preserving int/float
    if data is not None:
        if np.issubdtype(data.dtype, np.integer) and data.dtype.itemsize > 2:
            warnings.warn("Downcasting integer data to uint16.", stacklevel=2)
            data = data.astype(np.uint16)  # type: ignore[assignment]
        elif np.issubdtype(data.dtype, np.floating) and data.dtype.itemsize > 4:
            data = data.astype(np.float32)  # type: ignore[assignment]
    return data


def _downcast_and_downsample(
    data: T, three_d: bool, warn: bool = True
) -> tuple[T, tuple[int, ...]]:
    """Downcast >32bit data to 32bit, and downsample GPU texture limits are exceeded."""
    data = _downcast(data)
    downsample_factors: tuple[int, ...] = ()
    if data is not None:
        maxd = get_max_texture_sizes()[1 if three_d else 0]
        if maxd is not None:
            data, downsample_factors = downsample_data(data, maxd, warn=warn)  # type: ignore[assignment]
    return data, downsample_factors



"""Fast and flexible n-dimensional data viewer."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("ndv")
except PackageNotFoundError:
    __version__ = "uninstalled"

from . import data
from .controllers import ArrayViewer
from .models import DataWrapper
from .util import imshow
from .views import (
    call_later,
    process_events,
    run_app,
    set_canvas_backend,
    set_gui_backend,
)

__all__ = [
    "ArrayViewer",
    "DataWrapper",
    "call_later",
    "data",
    "imshow",
    "process_events",
    "run_app",
    "set_canvas_backend",
    "set_gui_backend",
]



from __future__ import annotations

import numpy as np

from ndv.views._util import downsample_data


def test_no_downsample_when_within_limit() -> None:
    data = np.zeros((50, 60), dtype=np.float32)
    result, factors = downsample_data(data, 64)
    assert factors == (1, 1)
    assert result is data  # no copy, same object


def test_exact_boundary_no_downsample() -> None:
    data = np.zeros((64, 64, 64), dtype=np.float32)
    result, factors = downsample_data(data, 64)
    assert factors == (1, 1, 1)
    assert result is data


def test_single_axis_overflow() -> None:
    data = np.zeros((10, 100), dtype=np.float32)
    result, factors = downsample_data(data, 64)
    assert factors == (1, 2)
    assert result.shape == (10, 50)


def test_all_axes_overflow() -> None:
    data = np.zeros((200, 150, 130), dtype=np.float32)
    result, factors = downsample_data(data, 64)
    assert factors == (4, 3, 3)
    assert result.shape == (50, 50, 44)


def test_2d_input() -> None:
    data = np.zeros((100, 80), dtype=np.float32)
    result, factors = downsample_data(data, 64)
    assert factors == (2, 2)
    assert result.shape == (50, 40)


def test_4d_input() -> None:
    data = np.zeros((10, 100, 80, 3), dtype=np.float32)
    result, factors = downsample_data(data, 64)
    assert len(factors) == 4
    assert factors == (1, 2, 2, 1)
    assert result.shape == (10, 50, 40, 3)


def test_returns_view_not_copy() -> None:
    """Downsampled result should be a view (no memory copy)."""
    data = np.ones((100, 100), dtype=np.float32)
    result, _ = downsample_data(data, 64)
    assert result.base is data


def test_max_size_1() -> None:
    data = np.zeros((5, 3), dtype=np.float32)
    result, factors = downsample_data(data, 1)
    assert factors == (5, 3)
    assert result.shape == (1, 1)



# /// script
# dependencies = [
#     "ndv[pyqt,vispy]",
#     "pyopencl",
#     "siphash24",
# ]
# ///
from __future__ import annotations

try:
    import pyopencl as cl
    import pyopencl.array as cl_array
except ImportError:
    raise ImportError("Please install pyopencl to run this example")
import ndv

# Set up OpenCL context and queue
context = cl.create_some_context(interactive=False)
queue = cl.CommandQueue(context)


gpu_data = cl_array.to_device(queue, ndv.data.nd_sine_wave())

ndv.imshow(gpu_data)



from abc import ABC, abstractmethod
from typing import Any


class Viewable(ABC):
    """ABC representing anything that can be viewed on screen.

    For example, a widget, a window, a frame, canvas, etc.
    """

    @abstractmethod
    def frontend_widget(self) -> Any:
        """Return the native object backing the viewable objects."""

    @abstractmethod
    def set_visible(self, visible: bool) -> None:
        """Sets the visibility of the view/widget itself."""

    @abstractmethod
    def close(self) -> None:
        """Close the view/widget."""



# /// script
# dependencies = [
#     "ndv[pyqt,vispy]",
#     "tensorstore",
# ]
# ///
from __future__ import annotations

import ndv

try:
    from ndv.data import cosem_dataset

    ts_array = cosem_dataset()
except ImportError:
    raise ImportError("Please install tensorstore to run this example")

ndv.imshow(ts_array)



from typing import ClassVar

from psygnal import SignalGroupDescriptor
from pydantic import BaseModel, ConfigDict


class NDVModel(BaseModel):
    """Base evented model for NDV models.

    Uses [pydantic.BaseModel][] and [psygnal.SignalGroupDescriptor][].
    """

    model_config = ConfigDict(
        validate_assignment=True,
        validate_default=True,
        extra="forbid",
    )
    events: ClassVar[SignalGroupDescriptor] = SignalGroupDescriptor()
